import os
import copy
from abc import abstractmethod
from inspect import isclass
from typing import List, Type, Union, Optional, Dict, Tuple, Any

from janis_core import translations
from janis_core.utils import first_value
from janis_core.utils.metadata import WorkflowMetadata
from janis_core.utils.logger import Logger
from janis_core.graph.node import Node, NodeTypes
from janis_core.graph.steptaginput import StepTagInput
from janis_core.translations import ExportPathKeywords
from janis_core.types import (
    DataType,
    ParseableType,
    get_instantiated_type,
    Array,
    InputSelector,
    Filename,
)
from janis_core.tool.tool import Tool, ToolType, ToolTypes, TInput, TOutput
from janis_core.tool.commandtool import CommandTool
from janis_core.types.data_types import is_python_primitive
from janis_core.utils.scatter import ScatterDescription
from janis_core.utils.validators import Validators

ConnectionSource = Union[Node, Tuple[Node, str]]


def verify_or_try_get_source(source: Union[ConnectionSource, List[ConnectionSource]]):
    if isinstance(source, list):
        return [verify_or_try_get_source(s) for s in source]
    node, tag = None, None
    if isinstance(source, tuple):
        node, tag = source
    else:
        node = source

    outs = node.outputs()
    if tag is None:
        if len(outs) > 1:
            raise Exception(
                f"Too many outputs of {node.id()} to guess the correct output"
            )
        tag = list(outs.keys())[0]

    if tag not in outs:
        tags = ", ".join([f"out.{o}" for o in outs.keys()])
        raise Exception(
            f"Couldn't find tag '{tag}' in outputs of '{node.id()}', "
            f"expected one of {tags}"
        )

    return node, tag


class InputNode(Node):
    def __init__(
        self,
        wf,
        identifier: str,
        datatype: DataType,
        default: any,
        value: any,
        doc: str = None,
    ):
        super().__init__(wf, NodeTypes.INPUT, identifier)
        self.datatype = datatype
        self.default = default
        self.doc = doc
        self.value = value

    def outputs(self) -> Dict[str, TOutput]:
        # Program will just grab first value anyway
        return {None: TOutput(self.identifier, self.datatype)}

    def inputs(self):
        return None


class StepNode(Node):
    def __init__(
        self,
        wf,
        identifier,
        tool: Tool,
        doc: str = None,
        scatter: ScatterDescription = None,
    ):
        super().__init__(wf, NodeTypes.STEP, identifier)
        self.tool = tool
        self.doc = doc
        self.scatter = scatter

    def inputs(self):
        return self.tool.inputs_map()

    def outputs(self):
        return self.tool.outputs_map()

    def _add_edge(self, tag: str, source: ConnectionSource):
        node, outtag = verify_or_try_get_source(source)

        if tag not in self.sources:
            self.sources[tag] = StepTagInput(self, tag)

        # If tag is in scatter.fields, then we can
        scatter = self.scatter and tag in self.scatter.fields

        return self.sources[tag].add_source(node, outtag, should_scatter=scatter)

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]

        return self.get_item(item)

    def __getitem__(self, item):
        return self.get_item(item)

    def get_item(self, item):
        ins = self.inputs()
        if item in ins:
            return self, item
        outs = self.outputs()
        if item in outs:
            return self, item

        tags = ", ".join(
            [f"in.{i}" for i in ins.keys()] + [f"out.{o}" for o in outs.keys()]
        )

        raise KeyError(
            f"Step '{self.id()}' with tool '{self.tool.id()}' has no identifier '{item}' ({tags})"
        )

    # "always set attributes keys"
    always_set = {"tool", "sources", "node_type", "identifier"}

    def __setitem__(self, key, value):
        if key not in StepNode.always_set and key in self.inputs():
            if isinstance(value, list):
                return [self._add_edge(key, v) for v in value]
            return self._add_edge(key, value)

        self.__dict__[key] = value


class OutputNode(Node):
    def __init__(
        self,
        wf,
        identifier: str,
        datatype: DataType,
        source: ConnectionSource,
        doc: str = None,
        output_tag: Union[str, InputSelector] = None,
        output_prefix: Union[
            str, InputSelector, List[Union[str, InputSelector]]
        ] = None,
    ):
        super().__init__(wf, NodeTypes.OUTPUT, identifier)
        self.datatype = datatype

        if source[0].node_type != NodeTypes.STEP:
            raise Exception(
                f"Unsupported connection type: {NodeTypes.OUTPUT} → {source[0].node_type}"
            )

        stype = source[0].outputs()[source[1]].outtype
        snode = source[0]
        if isinstance(snode, StepNode) and snode.scatter:
            stype = Array(stype)

        if not datatype.can_receive_from(stype):
            Logger.critical(
                f"Mismatch of types when joining to output node '{source[0].id()}.{source[1]}' to '{identifier}' "
                f"({stype.id()} -/→ {datatype.id()})"
            )

        self.source = verify_or_try_get_source(source)
        self.doc = doc
        self.output_tag = output_tag
        self.output_prefix = output_prefix

    def inputs(self) -> Dict[str, TInput]:
        # Program will just grab first value anyway
        return {None: TInput(self.identifier, self.datatype)}

    def outputs(self):
        return None


class Workflow(Tool):
    def __init__(self, **connections):
        super().__init__(metadata_class=WorkflowMetadata)

        self.connections = connections

        Logger.log(f"Creating workflow with identifier: '{self.id()}'")

        if not Validators.validate_identifier(self.id()):
            raise Exception(
                f"The identifier '{self.id()}' was invalid because {Validators.reason_for_failure(self.id())}"
            )

        # The following variables allow us to quickly check data about the graph
        self.nodes: Dict[str, Node] = {}

        self.input_nodes: Dict[str, InputNode] = {}
        self.step_nodes: Dict[str, StepNode] = {}
        self.output_nodes: Dict[str, OutputNode] = {}

        # Flags for different requirements that a workflow might need
        self.has_scatter = False
        self.has_subworkflow = False
        self.has_multiple_inputs = False

        # Now that we've initialised everything, we can "construct" the workflows for that subclass this class
        # else, for the WorkflowBuilder it will do nothing and they'll add workflows later
        self.constructor()

    @abstractmethod
    def friendly_name(self):
        pass

    @abstractmethod
    def constructor(self):
        """
        A place to construct your workflows. This is called directly after initialisation.
        :return:
        """
        pass

    def verify_identifier(self, identifier: str, component: str):

        if identifier in self.__dict__:
            raise Exception(
                f"'{identifier}' is a protected keyword for a janis workflow"
            )

        if identifier in self.nodes:
            existing = self.nodes[identifier]
            raise Exception(
                f"There already exists a node (and component) with id '{identifier}'. The added "
                f"component ('{component}') clashes with '{repr(existing)}')."
            )

        if not Validators.validate_identifier(identifier):
            raise Exception(
                f"The identifier '{identifier}' was invalid because {Validators.reason_for_failure(identifier)}"
            )

    def input(
        self,
        identifier: str,
        datatype: ParseableType,
        default: any = None,
        value: any = None,
        doc: str = None,
    ):
        """
        Create an input node on a workflow
        :return:
        """

        self.verify_identifier(identifier, repr(datatype))

        datatype = get_instantiated_type(datatype)
        if default:
            datatype.optional = True

        inp = InputNode(
            self,
            identifier=identifier,
            datatype=datatype,
            default=default,
            doc=doc,
            value=value,
        )
        self.nodes[identifier] = inp
        self.input_nodes[identifier] = inp
        return inp

    def output(
        self,
        identifier: str,
        datatype: Optional[ParseableType] = None,
        source: Union[StepNode, ConnectionSource] = None,
        output_tag: Union[
            str,
            InputSelector,
            ConnectionSource,
            List[Union[str, InputSelector, ConnectionSource]],
        ] = None,
        output_prefix: Union[str, InputSelector, ConnectionSource] = None,
    ):
        """
        Create an output on a workflow

        :param identifier: The identifier for the output
        :param datatype: Optional data type of the output to check. This will be automatically inferred if not provided.
        :param source: The source of the output, must be an output to a step node
        :param output_tag: A janis annotation for grouping outputs by this value.  If a list is passed, it represents
        a structure of nested directories, the first element being the root directory.
        At most, one InputSelector can resolve to an array, and this behaviour is only defined if the output
        scattered source, and the number of elements is equal.
        :param output_prefix: Decides the prefix that an output will have, or acts as a map if the InputSelector
        resolves to an array with equal length to the number of shards (scatters). Any other behaviour is defined and
        may result in an unexpected termination.
        :return:
        """
        self.verify_identifier(identifier, repr(datatype))

        if source is None:
            raise Exception("Output source must not be 'None'")

        node, tag = verify_or_try_get_source(source)
        if not datatype:
            datatype: DataType = node.outputs()[tag].outtype.received_type()

            if isinstance(node, StepNode) and node.scatter:
                datatype = Array(datatype)

        if output_prefix:
            op = output_prefix if isinstance(output_prefix, list) else [output_prefix]
            output_prefix = self.verify_output_source_type(
                identifier, op, "output_prefix"
            )
        if output_tag:
            ot = output_tag if isinstance(output_tag, list) else [output_tag]
            output_tag = self.verify_output_source_type(identifier, ot, "output_tag")

        otp = OutputNode(
            self,
            identifier=identifier,
            datatype=get_instantiated_type(datatype),
            source=(node, tag),
            output_tag=output_tag,
            output_prefix=output_prefix,
        )
        self.nodes[identifier] = otp
        self.output_nodes[identifier] = otp
        return otp

    def all_input_keys(self):
        from janis_core.translations.translationbase import TranslatorBase

        return super().all_input_keys() + list(
            TranslatorBase.build_resources_input(workflow=self, hints={}).keys()
        )

    def verify_output_source_type(
        self,
        identifier,
        out: Union[
            str,
            InputSelector,
            ConnectionSource,
            List[Union[str, InputSelector, ConnectionSource]],
        ],
        outtype: str,
    ):
        if isinstance(out, list):
            return [self.verify_output_source_type(identifier, o, outtype) for o in out]

        if isinstance(out, str):
            return out

        if isinstance(out, tuple):
            # ConnectionSource tuple
            out = out[0]

        if isinstance(out, Node):
            if not isinstance(out, InputNode):
                raise Exception(
                    f"The source for the {outtype} '{identifier}' was a {out.__class__.__name__} and must be an Input"
                )

            return InputSelector(out.identifier)

        if isinstance(out, InputSelector):
            keys = set(self.input_nodes.keys())
            if out.input_to_select not in keys:
                raise Exception(
                    f"Couldn't find the input {out.input_to_select} in the workflow, expected one of: "
                    + ", ".join(keys)
                )
            return out

        raise Exception(f"Invalidate type for {outtype}: {out.__class__.__name__}")

    def step(
        self,
        identifier: str,
        tool: Tool,
        scatter: Union[str, ScatterDescription] = None,
        ignore_missing=False,
    ):
        """
        Construct a step on this workflow.

        :param identifier: The identifier of the step, unique within the workflow.
        :param tool: The tool that should run for this step.
        :param scatter: Indicate whether a scatter should occur, on what, and how.
        :type scatter: Union[str, ScatterDescription]
        :param ignore_missing: Don't throw an error if required params are missing from this function
        :return:
        """

        self.verify_identifier(identifier, tool.id())

        if scatter is not None and isinstance(scatter, str):

            scatter = ScatterDescription(
                [scatter] if isinstance(scatter, str) else scatter
            )

        # verify scatter
        if scatter:
            ins = set(tool.inputs_map().keys())
            fields = set(scatter.fields)
            if any(f not in ins for f in fields):
                # if there is a field not in the input map, we have a problem
                extra_keys = ", ".join(f"'{f}'" for f in (fields - ins))
                raise Exception(
                    f"Couldn't scatter the field(s) for step '{identifier}' "
                    f"(tool: '{tool}') {extra_keys} as they were not found in the input map"
                )

        tool.workflow = self
        inputs = tool.inputs_map()

        connections = tool.connections

        provided_keys = set(connections.keys())
        all_keys = set(inputs.keys())
        required_keys = set(
            # The input is optional if it's optional or has default)
            i
            for i, v in inputs.items()
            if not (v.intype.optional or v.default is not None)
        )

        if not provided_keys.issubset(all_keys):
            unrecparams = ", ".join(provided_keys - all_keys)

            tags = ", ".join([f"in.{i}" for i in all_keys])

            raise Exception(
                f"Unrecognised parameters {unrecparams} when creating '{identifier}' ({tool.id()}). "
                f"Expected types: {tags}"
            )

        if not ignore_missing and not required_keys.issubset(provided_keys):
            missing = ", ".join(f"'{i}'" for i in (required_keys - provided_keys))
            raise Exception(
                f"Missing the parameters {missing} when creating '{identifier}' ({tool.id()})"
            )

        stp = StepNode(self, identifier=identifier, tool=tool, scatter=scatter)

        added_edges = []
        for (k, v) in connections.items():

            isfilename = isinstance(v, Filename)
            if is_python_primitive(v) or isfilename:
                inp_identifier = f"{identifier}_{k}"
                referencedtype = copy.copy(inputs[k].intype) if not isfilename else v
                parsed_type = get_instantiated_type(v)

                if parsed_type and not referencedtype.can_receive_from(parsed_type):
                    raise TypeError(
                        f"The type {parsed_type.id()} inferred from the value '{v}' is not "
                        f"compatible with the '{identifier}.{k}' type: {referencedtype.id()}"
                    )

                referencedtype.optional = True

                v = self.input(
                    inp_identifier,
                    referencedtype,
                    default=v.generated_filename() if isfilename else v,
                )
            if v is None:
                inp_identifier = f"{identifier}_{k}"
                v = self.input(inp_identifier, inputs[k].intype, default=v)

            verifiedsource = verify_or_try_get_source(v)
            if isinstance(verifiedsource, list):
                for vv in verifiedsource:
                    added_edges.append(stp._add_edge(k, vv))
            else:
                added_edges.append(stp._add_edge(k, verifiedsource))

        for e in added_edges:

            si = e.finish.sources[e.ftag] if e.ftag else first_value(e.finish.sources)
            self.has_multiple_inputs = self.has_multiple_inputs or si.multiple_inputs

        self.has_scatter = self.has_scatter or scatter is not None
        self.has_subworkflow = self.has_subworkflow or isinstance(tool, Workflow)
        self.nodes[identifier] = stp
        self.step_nodes[identifier] = stp

        return stp

    def __getattr__(self, item):
        if item in self.__dict__ or item == "nodes":
            return self.__dict__.get(item)

        if item in self.nodes:
            return self.nodes[item]

        raise AttributeError(
            f"AttributeError: '{type(self).__name__}' object has no attribute '{item}'"
        )

    def __getitem__(self, item):

        if item in self.nodes:
            return self.nodes[item]

        raise KeyError(f"KeyError: '{type(self).__name__}' object has no node '{item}'")

    @classmethod
    def type(cls) -> ToolType:
        return ToolTypes.Workflow

    def tool_inputs(self) -> List[TInput]:
        """
        List of ToolInputs of the workflow, we can toss out most of the metadata
        about positioning, prefixes, etc that the ToolInput class uses
        """
        return [
            TInput(i.id(), i.datatype, default=i.default)
            for i in self.input_nodes.values()
        ]

    def tool_outputs(self) -> List[TOutput]:
        """
        Similar to inputs, return a list of ToolOutputs of the workflow
        """
        return [TOutput(i.id(), i.datatype) for i in self.output_nodes.values()]

    @staticmethod
    def version():
        return "v0.1.0"

    def translate(
        self,
        translation: translations.SupportedTranslation,
        to_console=True,
        tool_to_console=False,
        to_disk=False,
        write_inputs_file=True,
        with_docker=True,
        with_hints=False,
        with_resource_overrides=False,
        validate=False,
        should_zip=True,
        export_path=ExportPathKeywords.default,
        merge_resources=False,
        hints=None,
        allow_null_if_not_optional=True,
        additional_inputs: Dict = None,
        max_cores=None,
        max_mem=None,
    ):
        return translations.translate_workflow(
            self,
            translation=translation,
            to_console=to_console,
            tool_to_console=tool_to_console,
            to_disk=to_disk,
            with_docker=with_docker,
            with_resource_overrides=with_resource_overrides,
            should_zip=should_zip,
            export_path=export_path,
            write_inputs_file=write_inputs_file,
            should_validate=validate,
            merge_resources=merge_resources,
            hints=hints,
            allow_null_if_not_optional=allow_null_if_not_optional,
            additional_inputs=additional_inputs,
            max_cores=max_cores,
            max_mem=max_mem,
        )

    def generate_inputs_override(
        self, additional_inputs=None, with_resource_overrides=False, hints=None
    ):
        """
        Generate the overrides to be used with Janis. Although it may work with
        other
        :return:
        """
        ad = additional_inputs or {}

        d = {
            i.id(): ad.get(i.id(), i.value or i.default)
            for i in self.input_nodes.values()
            if i.id() in ad or i.default or i.value or not i.datatype.optional
        }

        if with_resource_overrides:
            d.update(translations.CwlTranslator().build_resources_input(self, hints))

        return d

    def generate_resources_file(
        self,
        translation: translations.SupportedTranslation,
        hints: Dict[str, Any] = None,
        to_console=True,
    ):
        tr = translations.build_resources_input(self, translation, hints)
        if to_console:
            print(tr)
        return tr

    def get_tools(self) -> Dict[str, CommandTool]:
        tools: Dict[str, CommandTool] = {}
        for t in self._steps:
            tl = t.step.tool()
            if isinstance(tl, Workflow):
                tools.update(tl.get_tools())
            elif t.id() not in tools:
                tools[tl.id()] = tl
        return tools

    def report(self, to_console=True, tabulate_tablefmt=None):
        import tabulate

        tools = self.get_tools()
        keys = sorted(tools.keys(), key=lambda a: a[0].lower())

        header = ["tool", "version", "container"]
        data = []
        for t in keys:
            tool = tools[t]
            data.append(
                [
                    f"{tool.friendly_name()} ({tool.id()})",
                    tool.version(),
                    tool.container(),
                ]
            )

        retval = tabulate.tabulate(data, headers=header, tablefmt=tabulate_tablefmt)
        if to_console:
            print(retval)

        return retval

    def generate_resources_table(
        self,
        hints: Dict[str, Any],
        to_console=True,
        to_disk=False,
        output_type: str = "tsv",
    ):
        delim = "\t" if output_type == "tsv" else ","

        tools = self.get_tools()
        header = ["name", "cpu", "memory (GB)"]
        data = []

        for t in sorted(tools.keys()):
            tool = tools[t]
            data.append([tool.id(), tool.cpus(hints), tool.memory(hints)])

        data.sort(key=lambda a: a[0].lower())

        data.insert(0, header)

        if to_console:
            import tabulate

            print(tabulate.tabulate(data, headers="firstrow"))
        if to_disk:
            import csv

            d = ExportPathKeywords.resolve(
                ExportPathKeywords.default_no_spec, None, self.id()
            )
            path = d + f"resources.{output_type}"

            if not os.path.isdir(d):
                os.makedirs(d)

            Logger.info(f"Writing resources {output_type} to '{path}'")
            with open(path, "w+") as mf:
                writer = csv.writer(mf, delimiter=delim)
                for row in data:
                    writer.writerow(row)

        return data


class WorkflowBuilder(Workflow):
    def __init__(self, identifier: str = None, friendly_name: str = None):
        self._identifier = identifier
        self._name = friendly_name

        super().__init__()

    def id(self):
        return self._identifier

    def friendly_name(self):
        return self._name

    def __call__(self, **connections):
        self.connections = connections
        return self

    def constructor(self):
        """
        Empty placeholder as users will construct their workflow manually, and not as part of this class
        :return:
        """
        return self


if __name__ == "__main__":
    # from janis_unix import Echo

    wf = WorkflowBuilder("simple")

    from janis_bioinformatics.tools.cutadapt import CutAdapt_1_18 as Cutadapt
    from janis_bioinformatics.tools.common import BwaMem_SamToolsView
    from janis_bioinformatics.tools.gatk4 import Gatk4SortSam_4_0
    from janis_bioinformatics.data_types import FastaWithDict, Fastq

    # Inputs
    wf.input("sampleName", str)
    wf.input("reference", FastaWithDict)
    wf.input("fastq", Fastq)

    # Steps
    wf.step("cutadapt", Cutadapt(fastq=wf.fastq))
    wf.step(
        "bwamem",
        BwaMem_SamToolsView(
            reads=wf.cutadapt.out, sampleName=wf.sampleName, reference=wf.reference
        ),
    )
    wf.step(
        "sortsam",
        Gatk4SortSam_4_0(
            bam=wf.bwamem.out,
            sortOrder="coordinate",
            createIndex=True,
            validationStringency="SILENT",
            maxRecordsInRam=5000000,
            tmpDir=".",
        ),
    )

    # outputs
    wf.output("out", source=wf.sortsam)

    wf.translate("wdl")



class Resources:
    def __init__(self, cores_min=None, cores_max=None, ram_min=None, ram_max=None):
        self.coresMin = cores_min
        self.coresMax = cores_max
        self.ramMin = ram_min
        self.ramMax = ram_max



import warnings


class PipelineTranslatorException(Exception):
    pass


class DuplicateLabelIdentifier(Exception):
    pass


class InvalidNodeIdentifier(Exception):
    pass


class NodeNotFound(Exception):
    pass


class NotFoundException(Exception):
    pass


class InvalidInputsException(Exception):
    pass


class InvalidStepsException(Exception):
    pass


class TooManyArgsException(Exception):
    pass


class IncorrectArgsException(Exception):
    pass


class InvalidByProductException(Exception):
    pass


class ConflictingArgumentsException(Exception):
    pass


def deprecated(message):
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn(
                "{} is a deprecated function. {}".format(func.__name__, message),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator



import json
from _sha1 import sha1
from datetime import date


class Metadata(object):

    DATE_FORMAT = "%Y-%m-%d"

    def __init__(
        self,
        contributors=None,
        dateCreated=None,
        dateUpdated=None,
        institution=None,
        doi=None,
        citation=None,
        keywords=None,
        documentationUrl=None,
        documentation=None,
        short_documentation=None,
        version=None,
    ):
        """

        :param contributors:
        :param dateCreated:
        :param dateUpdated:
        :param institution:
        :param doi:
        :param citation:
        :param keywords:
        :param documentationUrl:
        :param documentation:
        :param short_documentation:
        :param version:
        """
        self.contributors = []
        if contributors:
            self.contributors = (
                contributors
                if isinstance(contributors, list)
                else contributors.split(",")
            )

        self.dateCreated = (
            dateCreated.strftime(self.DATE_FORMAT)
            if isinstance(dateCreated, date)
            else dateCreated
        )
        self.dateUpdated = (
            dateUpdated.strftime(self.DATE_FORMAT)
            if isinstance(dateUpdated, date)
            else dateUpdated
        )
        self.institution = institution
        self.doi = doi
        self.citation = citation
        self.keywords = keywords
        self.documentation = documentation
        self.short_documentation = short_documentation
        self.documentationUrl = documentationUrl
        self.version = version

    def update(self, **kwargs):
        for k in kwargs:
            self.__setattr__(k, kwargs[k])
        return self

    def get_dict(self, object_to_checksum):

        checksum = sha1(
            json.dumps(object_to_checksum, sort_keys=True).encode("utf-8")
        ).hexdigest()
        # https://stackoverflow.com/q/5884066

        d = {k: v for k, v in vars(self).items() if v is not None}
        d["checksum"] = checksum
        d["dateGenerated"] = date.today().strftime(self.DATE_FORMAT)

        return d


class WorkflowMetadata(Metadata):
    def get_dict(self, object_to_checksum, ninputs=None, nsteps=None, noutputs=None):
        d = super(WorkflowMetadata, self).get_dict(object_to_checksum)

        if ninputs:
            d["numberOfInputs"] = ninputs
        if nsteps:
            d["numberOfSteps"] = nsteps
        if noutputs:
            d["numberOfOutputs"] = noutputs

        return d


class ToolMetadata(Metadata):
    pass



from typing import List, Type, Optional
import pkg_resources, sys
from inspect import isfunction, ismodule, isabstract, isclass

from janis_core.tool.tool import Tool, ToolTypes
from janis_core.types.data_types import DataType
from janis_core.utils.logger import Logger, LogLevel
import janis_core.registry.entrypoints as EP
from janis_core.registry.registry import TaggedRegistry, Registry


class JanisShed:
    _byclassname = Registry()
    _toolshed = TaggedRegistry("latest")
    _typeshed = Registry()

    _has_been_hydrated = False

    # getters

    @staticmethod
    def get_by_class_name(name: str):
        JanisShed.hydrate()
        return JanisShed._byclassname.get(name)

    @staticmethod
    def get_tool(tool: str, version: str = None):
        JanisShed.hydrate()
        if version:
            version = version.lower()
        return JanisShed._toolshed.get(tool.lower(), version)

    @staticmethod
    def get_datatype(datatype: str):
        JanisShed.hydrate()
        return JanisShed._typeshed.get(datatype.lower())

    @staticmethod
    def get_all_tools() -> List[List[Tool]]:
        JanisShed.hydrate()
        return JanisShed._toolshed.objects()

    @staticmethod
    def get_all_datatypes() -> List[Type[DataType]]:
        JanisShed.hydrate()
        return JanisShed._typeshed.objects()

    # setters

    @staticmethod
    def add_tool(tool: Tool) -> bool:
        v: Optional[str] = tool.version()
        if not v:
            t = f"The tool {tool.id()} did not have a version and will not be registered"
            Logger.critical(t)
            return False

        JanisShed._byclassname.register(tool.__class__.__name__, tool)
        return JanisShed._toolshed.register(tool.id().lower(), v.lower(), tool)

    @staticmethod
    def add_type(datatype: Type[DataType]) -> bool:
        JanisShed._byclassname.register(datatype.__name__, datatype)
        return JanisShed._typeshed.register(datatype.name().lower(), datatype)

    @staticmethod
    def hydrate(force=False, modules: list = None):
        # go get everything
        if JanisShed._has_been_hydrated and not force:
            return

        if not modules:
            modules = []
            modules.extend(JanisShed._get_datatype_entrypoints())
            modules.extend(JanisShed._get_tool_entrypoints())

        Logger.log("Setting CONSOLE_LEVEL to None while traversing modules")
        cl = Logger.CONSOLE_LEVEL
        Logger.set_console_level(None)
        seen_modules = set()
        seen_classes = set()
        for m in modules:
            JanisShed.traverse_module(
                m, seen_modules=seen_modules, seen_classes=seen_classes
            )
        Logger.set_console_level(cl)
        Logger.log(
            f"Restoring CONSOLE_LEVEL to {LogLevel.get_str(cl)} now that Janis shed has been hydrated"
        )

        JanisShed._has_been_hydrated = True

    @staticmethod
    def _get_datatype_entrypoints():
        ep = []
        eps = pkg_resources.iter_entry_points(group=EP.DATATYPES)
        for entrypoint in eps:
            try:
                m = entrypoint.load()
                ep.append(m)
            except ImportError as e:
                t = f"Couldn't import janis data_type extension '{entrypoint.name}': {e}"
                Logger.critical(t)
                continue
        return ep

    @staticmethod
    def _get_tool_entrypoints():
        ep = []
        eps = pkg_resources.iter_entry_points(group=EP.TOOLS)
        for entrypoint in eps:
            try:
                m = entrypoint.load()
                ep.append(m)
            except ImportError as e:
                t = f"Couldn't import janis data_type extension '{entrypoint.name}': {e}"
                Logger.critical(t)
                continue
        return ep

    @staticmethod
    def traverse_module(module, seen_modules: set, seen_classes: set):
        if module.__name__ in seen_modules:
            return

        seen_modules.add(module.__name__)

        q = {
            n: cls
            for n, cls in list(module.__dict__.items())
            if not n.startswith("__")
            and type(cls) != type
            and not (ismodule(cls) and cls.__name__ in seen_modules)
            and not (isclass(cls) and cls in seen_classes)
        }

        for k in q:
            cls = q[k]
            JanisShed.process_cls(cls, seen_modules, seen_classes)

    @staticmethod
    def process_cls(cls, seen_modules, seen_classes: set):
        try:
            if ismodule(cls):
                return JanisShed.traverse_module(cls, seen_modules, seen_classes)
            elif isfunction(cls) or isabstract(cls):
                return
            elif not isclass(cls):
                return

            seen_classes.add(cls)

            if issubclass(cls, DataType):
                return JanisShed.add_type(cls)
            elif not hasattr(cls, "type") or not callable(cls.type):
                return
            elif cls.type() == ToolTypes.Workflow:
                return JanisShed.add_tool(cls())
            elif cls.type() == ToolTypes.CommandTool:
                return JanisShed.add_tool(cls())

        except Exception as e:
            Logger.log(f"{str(e)} for type {type(cls)}")



from enum import Enum
from typing import List

ScatterMethod = str


class ScatterMethods(Enum):
    """
    The scatter methods that Janis supports:
        - Dot: Inner product of two vectors of the same length, eg: Dot [A, B, C] · [1, 2, 3] => [A1, B2, C3]
        - Cross: Cartesian product of two vectors, producing every combination of the scattered inputs [A, B] x [1, 2] => [A1, A2, B1, B2]
    """

    dot: ScatterMethod = "dot"
    cross: ScatterMethod = "cross"

    def cwl(self):
        if self == ScatterMethods.dot:
            return "dotproduct"
        elif self == ScatterMethods.cross:
            return "flat_crossproduct"  # "nested_crossproduct"

        raise Exception(f"Unrecognised scatter method: '{self.value}'")


class ScatterDescription:
    """
    Class for keeping track of scatter information
    """

    def __init__(self, fields: List[str], method: ScatterMethods = None):
        """

        :param fields: The fields of the the tool that should be scattered on.
        :param method: The method that should be used to scatter the two arrays
        :type method: ScatterMethods
        """
        self.fields = fields
        self.method: ScatterMethods = method
        if len(fields) > 1 and method is None:
            raise Exception(
                "When there is more than one field, a ScatterMethod must be selected"
            )



"""
This file contains all the information required to get all the tools and data_types for all
the extension available to Janis! It will put them in a "TaggedRegistry" object, where you can
access a tool by it's name (with optional version).

You should be able to get all the tools + versions.
"""


from typing import Dict, List, Generic, TypeVar, Optional

from janis_core.utils.logger import Logger


class RegistryException(Exception):
    pass


T = TypeVar("T")


class Registry(Generic[T]):
    def __init__(self):
        self.registry: Dict[str, T] = {}

    def register(self, name: str, obj: T):
        if name in self.registry:
            return False
        self.registry[name] = obj
        return True

    def objects(self) -> List[T]:
        return list(self.registry.values())

    def get(self, type_name) -> Optional[T]:
        return self.registry.get(type_name)

    def __contains__(self, item):
        return item in self.registry


class TaggedRegistry(Generic[T]):
    def __init__(self, default_tag):

        self.registry: Dict[str, Dict[str, T]] = {}
        self.default_tag = default_tag
        # We'll specifically make registry[name]["latest"] = (obj: T, version: Comparable)

    def tag_or_default(self, tag):
        return tag if tag is not None else self.default_tag

    def register(self, name: str, tag: Optional[str], obj: T):

        d = self.registry.get(name) if name in self.registry else {}

        # We handle the default tag differently, so make sure the tag isn't
        # the default tag either.
        if tag is not None and tag != self.default_tag:
            if tag in d:
                return False
            d[tag] = obj

            if self.default_tag in d:
                prev, prev_tag = d[self.default_tag]
                if prev_tag is None or tag > prev_tag:
                    d[self.default_tag] = (obj, tag)
            else:
                d[self.default_tag] = (obj, tag)

        elif self.default_tag not in d:
            d[self.default_tag] = (obj, None)

        self.registry[name] = d
        return True

    def objects(self) -> List[List[T]]:
        return list(
            [
                list(y for y in x.values() if not isinstance(y, tuple))
                for x in self.registry.values()
            ]
        )

    def get(self, type_name, tag: Optional[str]) -> Optional[T]:
        if type_name not in self.registry:
            return None
        tagged_objs = self.registry[type_name]

        if tag is None or tag == self.default_tag:
            if self.default_tag in tagged_objs:
                return tagged_objs.get(self.default_tag)[0]
            return None

        if tag not in tagged_objs:
            Logger.log(
                "Found collection '{tool}' in registry, but couldn't find tag '{tag}'".format(
                    tool=type_name, tag=tag
                )
            )
            return None

        return tagged_objs[tag]

    def __contains__(self, item):
        return item in self.registry



from typing import Optional, Dict, Any

from janis_core.graph.node import Node, NodeTypes
from janis_core.tool.tool import TInput, TOutput
from janis_core.types.common_data_types import Array
from janis_core.utils import first_value
from janis_core.utils.logger import Logger


def full_lbl(node: Node, tag: Optional[str]) -> str:
    if tag is None:
        return node.id()
    return f"{node.id()}/{tag}"


def full_dot(node: Node, tag: Optional[str]) -> str:
    if tag is None:
        return node.id()
    return f"{node.id()}.{tag}"


class Edge:
    def __init__(
        self,
        start: Node,
        stag: Optional[str],
        finish: Node,
        ftag: Optional[str],
        should_scatter,
    ):
        Logger.log(
            f"Creating edge: ({NodeTypes.to_str(start.node_type)}) '{start.id()}.{stag}' → "
            f"({NodeTypes.to_str(finish.node_type)}) '{finish.id()}.{ftag}'"
        )

        self.start: Node = start
        self.stag: Optional[str] = stag
        self.finish: Node = finish
        self.ftag: Optional[str] = ftag
        self.compatible_types: Optional[bool] = None
        self.scatter = should_scatter

        self.validate_tags()
        self.check_types()

    def source_slashed(self):
        return full_lbl(self.start, self.stag)

    def source_dotted(self):
        return full_dot(self.start, self.stag)

    def validate_tags(self):
        if (
            self.start.node_type == NodeTypes.STEP
            and self.stag not in self.start.outputs()
        ):
            raise Exception(
                f"Could not find the tag '{self.stag}' in the inputs of '{self.start.id()}'"
            )
        if (
            self.finish.node_type == NodeTypes.STEP
            and self.ftag not in self.finish.inputs()
        ):
            raise Exception(
                f"Could not find the tag '{self.ftag}' in the outputs of '{self.finish.id()}': {list(self.finish.inputs().keys())}"
            )

    def check_types(self):
        from janis_core.workflow.workflow import InputNode, StepNode

        stoolin: TOutput = self.start.outputs()[
            self.stag
        ] if self.stag is not None else first_value(self.start.outputs())
        ftoolin: TInput = self.finish.inputs()[
            self.ftag
        ] if self.ftag is not None else first_value(self.finish.inputs())

        stype = stoolin.outtype
        ftype = ftoolin.intype

        start_is_scattered = (
            isinstance(self.start, StepNode) and self.start.scatter is not None
        )

        if start_is_scattered:
            Logger.log(
                f"This edge merges the inputs from '{full_dot(self.start, self.stag)}' for "
                f"'{full_dot(self.finish, self.ftag)}'"
            )
            stype = Array(stype)

        if self.scatter:
            if not isinstance(stype, Array):
                raise Exception(
                    f"Scatter was required for '{self.start.id()}.{self.stag} → '{self.finish.id()}.{self.ftag}' but "
                    f"the input type was {type(stype).__name__} and not an array"
                )
            stype = stype.subtype()

        source_has_default = (
            isinstance(self.start, InputNode) and self.start.default is not None
        )

        # Scatters are handled automatically by the StepTagInput Array unwrapping
        # Merges are handled automatically by the `start_is_scattered` Array wrap

        self.compatible_types = ftype.can_receive_from(stype, source_has_default)
        if not self.compatible_types:
            if isinstance(ftype, Array) and ftype.subtype().can_receive_from(stype):
                self.compatible_types = True

        if not self.compatible_types:

            s = full_dot(self.start, self.stag)
            f = full_dot(self.finish, self.ftag)
            message = (
                f"Mismatch of types when joining '{s}' to '{f}': "
                f"{stoolin.outtype.id()} -/→ {ftoolin.intype.id()}"
            )
            if isinstance(stype, Array) and ftype.can_receive_from(stype.subtype()):
                message += " (did you forget to SCATTER?)"
            Logger.critical(message)


class StepTagInput:
    """
    This class represents the connections that a single input on a step has. Hence, a step
    will have one StepTagInput for each potential input of the tool.
    """

    def __init__(self, finish: Node, finish_tag: str):

        self.finish: Node = finish
        self.ftag: Optional[str] = finish_tag

        self.default = None
        self.multiple_inputs = False

        self.source_map: Dict[str, Edge] = {}

    def add_source(self, start: Node, stag: Optional[str], should_scatter) -> Edge:
        """
        Add a connection
        :param start:
        :param stag:
        :param should_scatter:
        :return:
        """

        from janis_core.workflow.workflow import StepNode

        stype = (
            start.outputs()[stag] if stag is not None else first_value(start.outputs())
        ).outtype
        ftype = (
            self.finish.inputs()[self.ftag]
            if self.ftag is not None
            else first_value(self.finish.inputs())
        ).intype

        start_is_scattered = isinstance(start, StepNode) and start.scatter is not None

        if start_is_scattered:
            Logger.log(
                f"This edge merges the inputs from '{full_dot(start, stag)}' for "
                f"'{full_dot(self.finish, self.ftag)}'"
            )
            stype = Array(stype)

        if should_scatter:
            if not isinstance(stype, Array):
                raise Exception(
                    f"Scatter was required for '{start.id()}.{stag} → '{self.finish.id()}.{self.ftag}' but "
                    f"the input type was {type(stype).__name__} and not an array"
                )
            stype = stype.subtype()

        if len(self.source_map) == 1 and start.id() not in self.source_map:
            self.multiple_inputs = True

            if not isinstance(ftype, Array):
                raise Exception(
                    f"Adding multiple inputs to '{self.finish.id()}' and '{ftype.id()}' is not an array"
                )

        if not isinstance(stype, Array) and isinstance(ftype, Array):
            # https://www.commonwl.org/user_guide/misc/#connect-a-solo-value-to-an-input-that-expects-an-array-of-that-type
            self.multiple_inputs = True

        e = Edge(start, stag, self.finish, self.ftag, should_scatter=should_scatter)
        self.source_map[start.id()] = e
        return e

    def set_default(self, default: Any):
        Logger.log(
            f"Setting the default of '{self.finish.id()}.{self.ftag}' to be '{str(default)}'"
        )
        self.default = default

    def source(self):
        n = len(self.source_map)
        if n == 0:
            return None
        elif n == 1:
            return first_value(self.source_map)
        else:
            return list(self.source_map.values())

    def dotted_source(self):
        n = len(self.source_map)

        if n == 0:
            return None
        elif n == 1:
            return first_value(self.source_map).source_dotted()
        else:
            return [e.source_dotted() for e in self.source_map.values()]

    def slashed_source(self):
        n = len(self.source_map)
        if n == 0:
            return None
        elif n == 1:
            return first_value(self.source_map).source_slashed()
        else:
            return [e.source_slashed() for e in self.source_map.values()]



from abc import ABC, abstractmethod
import os
from path import Path
from typing import Tuple, List, Dict, Optional, Any

from janis_core.tool.commandtool import ToolInput
from janis_core.types.common_data_types import Int
from janis_core.translations.exportpath import ExportPathKeywords
from janis_core.utils.logger import Logger

# from janis_core.workflow.input import Input, InputNode


class TranslatorBase(ABC):
    """
    So you're thinking about adding a new translation :)

    This class will hopefully give you a pretty good indication
    on what's required to add a new translation, however what I
    can't fully talk about are all the concepts.

    To add a new translation, you should understand at least these
    workflow and janis concepts:
        - inputs
        - outputs
        - steps + tools
        - secondary files
        - command line binding of tools
            - position, quoted
        - selectors
            - input selectors
            - wildcard glob selectors
            - cpu and memory selectors
        - propagated resource overrides

    I'd ask that you keep your function sizes small and direct,
    and then write unit tests to cover each component of the translation
    and then an integration test of the whole translation on the related workflows.

    You can find these in /janis/tests/test_translation_*.py)
    """

    def __init__(self, name):
        self.name = name

    def translate(
        self,
        workflow,
        to_console=True,
        tool_to_console=False,
        with_docker=True,
        with_resource_overrides=False,
        to_disk=False,
        write_inputs_file=True,
        export_path=ExportPathKeywords.default,
        should_validate=False,
        should_zip=True,
        merge_resources=False,
        hints=None,
        allow_null_if_not_optional=True,
        additional_inputs: Dict = None,
        max_cores=None,
        max_mem=None,
    ):

        # self.validate_inputs(workflow._inputs, allow_null_if_not_optional)

        tr_wf, tr_tools = self.translate_workflow(
            workflow,
            with_docker=with_docker,
            with_resource_overrides=with_resource_overrides,
        )
        tr_inp = self.build_inputs_file(
            workflow,
            recursive=False,
            merge_resources=merge_resources,
            hints=hints,
            additional_inputs=additional_inputs,
            max_cores=max_cores,
            max_mem=max_mem,
        )
        tr_res = self.build_resources_input(workflow, hints)

        str_wf = self.stringify_translated_workflow(tr_wf)
        str_inp = self.stringify_translated_inputs(tr_inp)
        str_tools = [
            (
                "tools/" + self.tool_filename(t),
                self.stringify_translated_workflow(tr_tools[t]),
            )
            for t in tr_tools
        ]
        str_resources = self.stringify_translated_inputs(tr_res)

        if to_console:
            print("=== WORKFLOW ===")
            print(str_wf)
            if tool_to_console:
                print("\n=== TOOLS ===")
                [print(f":: {t[0]} ::\n" + t[1]) for t in str_tools]
            print("\n=== INPUTS ===")
            print(str_inp)
            if not merge_resources and with_resource_overrides:
                print("\n=== RESOURCES ===")
                print(str_resources)

        d = ExportPathKeywords.resolve(
            export_path, workflow_spec=self.name, workflow_name=workflow.id()
        )

        fn_workflow = self.workflow_filename(workflow)
        fn_inputs = self.inputs_filename(workflow)
        fn_resources = self.resources_filename(workflow)

        if to_disk and write_inputs_file:
            if not os.path.isdir(d):
                os.makedirs(d)

            with open(os.path.join(d, fn_inputs), "w+") as f:
                Logger.log(f"Writing {fn_inputs} to disk")
                f.write(str_inp)
                Logger.log(f"Written {fn_inputs} to disk")
        else:
            Logger.log("Skipping writing input (yaml) job file")

        if to_disk:

            toolsdir = os.path.join(d, "tools")
            if not os.path.isdir(toolsdir):
                os.makedirs(toolsdir)

            Logger.info(f"Exporting workflow files to '{d}'")

            with open(os.path.join(d, fn_workflow), "w+") as wf:
                Logger.log(f"Writing {fn_workflow} to disk")
                wf.write(str_wf)
                Logger.log(f"Wrote {fn_workflow}  to disk")

            for (fn_tool, str_tool) in str_tools:
                with open(os.path.join(d, fn_tool), "w+") as toolfp:
                    Logger.log(f"Writing {fn_tool} to disk")
                    toolfp.write(str_tool)
                    Logger.log(f"Written {fn_tool} to disk")

            if not merge_resources and with_resource_overrides:
                print("\n=== RESOURCES ===")
                with open(os.path.join(d, fn_resources), "w+") as wf:
                    Logger.log(f"Writing {fn_resources} to disk")
                    wf.write(str_wf)
                    Logger.log(f"Wrote {fn_resources}  to disk")
                print(str_resources)

            import subprocess

            if should_zip:
                Logger.info("Zipping tools")
                with Path(d):
                    FNULL = open(os.devnull, "w")
                    zip_result = subprocess.run(
                        ["zip", "-r", "tools.zip", "tools/"], stdout=FNULL
                    )
                    if zip_result.returncode == 0:
                        Logger.info("Zipped tools")
                    else:
                        Logger.critical(zip_result.stderr)

            if should_validate:
                with Path(d):

                    Logger.info(f"Validating outputted {self.name}")

                    enved_vcs = [
                        (os.getenv(x[1:]) if x.startswith("$") else x)
                        for x in self.validate_command_for(
                            fn_workflow, fn_inputs, "tools/", "tools.zip"
                        )
                    ]

                    cwltool_result = subprocess.run(enved_vcs)
                    if cwltool_result.returncode == 0:
                        Logger.info(
                            "Exported workflow was validated by: " + " ".join(enved_vcs)
                        )
                    else:
                        Logger.critical(cwltool_result.stderr)

        return str_wf, str_inp, str_tools

    def translate_tool(
        self,
        tool,
        to_console=True,
        to_disk=False,
        export_path=None,
        with_docker=True,
        with_resource_overrides=False,
        max_cores=None,
        max_mem=None,
    ):

        tool_out = self.stringify_translated_tool(
            self.translate_tool_internal(
                tool,
                with_docker=with_docker,
                with_resource_overrides=with_resource_overrides,
            )
        )

        if to_console:
            print(tool_out)

        if to_disk:
            d = ExportPathKeywords.resolve(
                export_path, workflow_spec=self.name, workflow_name=tool.id()
            )
            if not os.path.exists(d):
                os.makedirs(d)
            fn_tool = self.tool_filename(tool)
            with open(os.path.join(d, fn_tool), "w+") as wf:
                Logger.log(f"Writing {fn_tool} to disk")
                wf.write(tool_out)
                Logger.log(f"Wrote {fn_tool}  to disk")

        return tool_out

    @classmethod
    def validate_inputs(cls, inputs, allow_null_if_optional):
        return True
        # invalid = [
        #     i
        #     for i in inputs
        #     if not i.input.validate_value(
        #         allow_null_if_not_optional=allow_null_if_optional
        #     )
        # ]
        # if len(invalid) == 0:
        #     return True
        # raise TypeError(
        #     "Couldn't validate inputs: "
        #     + ", ".join(
        #         f"{i.id()} (expected: {i.input.data_type.id()}, "
        #         f"got: '{TranslatorBase.get_type(i.input.value)}')"
        #         for i in invalid
        #     )
        # )

    @staticmethod
    def get_type(t):
        if isinstance(t, list):
            q = set(TranslatorBase.get_type(tt) for tt in t)
            if len(q) == 0:
                return "empty array"
            val = q.pop() if len(q) == 1 else "Union[" + ", ".join(q) + "]"
            return f"Array<{val}>"

        return type(t).__name__

    @classmethod
    @abstractmethod
    def translate_workflow(
        cls, workflow, with_docker=True, with_resource_overrides=False
    ) -> Tuple[any, Dict[str, any]]:
        pass

    @classmethod
    @abstractmethod
    def translate_tool_internal(
        cls, tool, with_docker=True, with_resource_overrides=False
    ):
        pass

    @classmethod
    @abstractmethod
    def build_inputs_file(
        cls,
        workflow,
        recursive=False,
        merge_resources=False,
        hints=None,
        additional_inputs: Dict = None,
        max_cores=None,
        max_mem=None,
    ) -> Dict[str, any]:
        pass

    @classmethod
    def build_resources_input(
        cls, workflow, hints, max_cores=None, max_mem=None, inputs=None, prefix=""
    ):
        from janis_core.workflow.workflow import Workflow, Tool, CommandTool

        # returns a list of key, value pairs
        steps: Dict[str, Optional[Any]] = {}
        inputs = inputs or {}

        # prefix = ""
        # if not prefix:
        #     prefix = ""
        # else:
        #     prefix += ""

        for s in workflow.step_nodes.values():
            tool: Tool = s.tool

            if isinstance(tool, CommandTool):
                tool_pre = prefix + s.id() + "_"
                cpus = inputs.get(f"{s.id()}_runtime_cpu", tool.cpus(hints) or 1)
                mem = inputs.get(f"{s.id()}_runtime_memory", tool.memory(hints))
                disk = inputs.get(f"{s.id()}_runtime_disks", "local-disk 60 SSD")

                if max_cores and cpus > max_cores:
                    Logger.info(
                        f"Tool '{tool.tool()}' exceeded ({cpus}) max number of cores ({max_cores}), "
                        "this was dropped to the new maximum"
                    )
                    cpus = max_cores
                if mem and max_mem and mem > max_mem:
                    Logger.info(
                        f"Tool '{tool.tool()}' exceeded ({mem} GB) max amount of memory({max_mem} GB), "
                        "this was dropped to the new maximum"
                    )
                    mem = max_mem
                steps.update(
                    {
                        tool_pre + "runtime_memory": mem,
                        tool_pre + "runtime_cpu": cpus,
                        tool_pre + "runtime_disks": disk,
                    }
                )
            elif isinstance(tool, Workflow):
                tool_pre = prefix + s.id() + "_"
                steps.update(
                    cls.build_resources_input(
                        tool,
                        hints,
                        max_cores=max_cores,
                        max_mem=max_mem,
                        prefix=tool_pre,
                        inputs={
                            k[len(s.id()) + 1 :]: v
                            for k, v in inputs.items()
                            if k.startswith(s.id())
                        },
                    )
                )

        return steps

    @staticmethod
    def inp_can_be_skipped(inp, override_value=None):
        return (
            inp.default is None
            and override_value is None
            # and not inp.include_in_inputs_file_if_none
            and (inp.datatype.optional and inp.default is None)
        )

    # Resource overrides
    @staticmethod
    def get_resource_override_inputs() -> List[ToolInput]:
        return [
            ToolInput("runtime_cpu", Int(optional=True), default=1),
            ToolInput("runtime_memory", Int(optional=True), default=4),
        ]

    # STRINGIFY

    @staticmethod
    @abstractmethod
    def stringify_translated_workflow(wf):
        pass

    @staticmethod
    @abstractmethod
    def stringify_translated_tool(tool):
        pass

    @staticmethod
    @abstractmethod
    def stringify_translated_inputs(inputs):
        pass

    # OUTPUTS

    @staticmethod
    @abstractmethod
    def workflow_filename(workflow):
        pass

    @staticmethod
    @abstractmethod
    def inputs_filename(workflow):
        pass

    @staticmethod
    @abstractmethod
    def tool_filename(tool):
        pass

    @staticmethod
    def dependencies_filename(workflow):
        return "tools.zip"

    @staticmethod
    @abstractmethod
    def resources_filename(workflow):
        pass

    # VALIDATION

    @staticmethod
    @abstractmethod
    def validate_command_for(wfpath, inppath, tools_dir_path, tools_zip_path):
        pass



"""
    Custom Types::

    We want to define some standard interface that allows for the implementation of types.
    Types must be supported as Inputs and Outputs, and must be comparable, ie we can compare
    the output of one step to another step. We'll guarantee that

    We'll also may require separate interfaces for CWL and WDL implementations, we may not be
    able to genericise the interface enough to support an automatic conversion

    We are allowed to require that a type must register itself when loaded.

"""
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Union

import cwlgen as cwl
from wdlgen import WdlType

from janis_core.utils import is_array_prefix
from janis_core.utils.logger import Logger

NativeType = str
PythonPrimitive = Union[str, float, int, bool]


def is_python_primitive(t):
    return (isinstance(t, list) and len(t) > 0 and is_python_primitive(t[0])) or any(
        isinstance(t, T) for T in [str, float, int, bool]
    )


# see below for ParseableType


class NativeTypes:
    kStr: NativeType = "str"
    kInt: NativeType = "int"
    kLong: NativeType = "long"
    kFloat: NativeType = "float"
    kBool: NativeType = "bool"
    kDouble: NativeType = "double"
    kFile: NativeType = "file"
    kDirectory: NativeType = "dir"
    kArray: NativeType = "array"
    kStdout: NativeType = "stdout"
    kStderr: NativeType = "stderr"

    _primitives: List[NativeType] = [kStr, kInt, kFloat, kLong, kDouble, kBool, kDouble]
    all: List[NativeType] = _primitives + [kFile, kDirectory, kArray, kStdout, kStderr]

    @staticmethod
    def is_primitive(t: NativeType) -> bool:
        return t in NativeTypes._primitives

    @staticmethod
    def is_valid(t: str):
        return t in NativeTypes.all

    @staticmethod
    def map_to_cwl(t: NativeType):
        if t == NativeTypes.kBool:
            return cwl.CwlTypes.BOOLEAN
        elif t == NativeTypes.kInt:
            return cwl.CwlTypes.INT
        elif t == NativeTypes.kLong:
            return cwl.CwlTypes.LONG
        elif t == NativeTypes.kFloat:
            return cwl.CwlTypes.FLOAT
        elif t == NativeTypes.kDouble:
            return cwl.CwlTypes.DOUBLE
        elif t == NativeTypes.kStr:
            return cwl.CwlTypes.STRING
        elif t == NativeTypes.kFile:
            return cwl.CwlTypes.FILE
        elif t == NativeTypes.kDirectory:
            return cwl.CwlTypes.DIRECTORY
        elif t == NativeTypes.kArray:
            return cwl.CwlTypes.ARRAY
        elif t == NativeTypes.kStdout:
            return cwl.CwlTypes.STDOUT
        elif t == NativeTypes.kStderr:
            return cwl.CwlTypes.STDERR
        raise Exception(
            f"Unhandled primitive type {t}, expected one of {', '.join(NativeTypes.all)}"
        )

    @staticmethod
    def map_to_wdl(t: NativeType):
        import wdlgen as wdl

        if t == NativeTypes.kBool:
            return wdl.PrimitiveType.kBoolean
        elif t == NativeTypes.kInt:
            return wdl.PrimitiveType.kInt

        elif (
            t == NativeTypes.kLong
            or t == NativeTypes.kFloat
            or t == NativeTypes.kDouble
        ):
            return wdl.PrimitiveType.kFloat
        elif t == NativeTypes.kStr:
            return wdl.PrimitiveType.kString
        elif t == NativeTypes.kFile:
            return wdl.PrimitiveType.kFile
        elif t == NativeTypes.kStdout:
            return wdl.PrimitiveType.kFile
        elif t == NativeTypes.kStderr:
            return wdl.PrimitiveType.kFile
        elif t == NativeTypes.kDirectory:
            Logger.log(
                "Using data_type 'Directory' for wdl, this requires cromwell>=37 and language=development"
            )
            return wdl.PrimitiveType.kDirectory
        elif t == NativeTypes.kArray:
            return wdl.ArrayType.kArray
        raise Exception(
            f"Unhandled primitive type {t}, expected one of {', '.join(NativeTypes.all)}"
        )

    @staticmethod
    def default_value(t: NativeType):
        if t == NativeTypes.kBool:
            return True
        elif t == NativeTypes.kInt:
            return 0
        elif t == NativeTypes.kLong:
            return 0.0
        elif t == NativeTypes.kFloat:
            return 0.0
        elif t == NativeTypes.kDouble:
            return 0.0
        elif t == NativeTypes.kStr:
            return "nothing"
        elif t == NativeTypes.kFile:
            return {"type": "File", "path": "path/to/file"}
        elif t == NativeTypes.kDirectory:
            return {"type": "Directory", "path": "path/to/file"}
        elif t == NativeTypes.kArray:
            return []


class DataType(ABC):
    def __init__(self, optional=False):
        self.optional = optional
        self.is_prim = NativeTypes.is_primitive(self.primitive())

    @staticmethod
    @abstractmethod
    def name() -> str:
        raise Exception("Subclass MUST override name field")

    @classmethod
    def __hash__(cls):
        return cls.name()

    @staticmethod
    def secondary_files() -> Optional[List[str]]:
        return None

    @staticmethod
    @abstractmethod
    def primitive() -> NativeType:
        raise Exception("Subclass MUST override the 'primitive' method")

    @staticmethod
    @abstractmethod
    def doc() -> str:
        """
        Subclasses should override this class to provide additional information on how to
        correctly provide data to the class, what inputs it may have and what other types
        are compatible
        """
        raise Exception("Subclass MUST override the 'doc' field")

    # The following methods don't need to be overriden, but can be

    def id(self):
        if self.optional:
            return f"Optional<{self.name()}>"
        return self.name()

    @abstractmethod
    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        pass

    @abstractmethod
    def invalid_value_hint(self, meta):
        pass

    def identify(self):
        print(self.id())

    def can_receive_from(self, other, source_has_default=False) -> bool:
        """
        Can this class receive from $other, likely going to be type(a) == type(b)
        :param other:
        :param source_has_default: If the source has default, then we can return true even if the source is optional
        :return:
        """

        # Depending on the way types are imported, the 'isinstance' method doesn't always work
        #
        # A small example:
        #     from janis import File as F1
        #     from janis_core import File as F2
        #
        # Although these are the same definition, they won't actually compare to the same value

        receive_from = list(
            reversed([x.__name__ for x in type(other.received_type()).mro()])
        )
        receive_to = list(
            reversed([x.__name__ for x in type(self.received_type()).mro()])
        )

        if not is_array_prefix(receive_to, receive_from):
            return False
        if self.optional or source_has_default:
            # If I'm optional I can receive from optional / non optional
            return True
        # If I'm not optional, I must receive from not optional
        return not other.optional

    def received_type(self):
        """
        The type that will be received if joined from this type
        :return: mostly self, except for STDOUT | STDERR | STDIN
        """
        return self

    def input_field_from_input(self, meta):
        """
        Method to convert the field definition into a generic CWL-esque response
        :param meta:
        :return:
        """
        return None

    def _question_mark_if_optional(self, has_default: bool = False):
        return "?" if self.optional or has_default else ""

    def cwl_type(self, has_default=False):
        tp = NativeTypes.map_to_cwl(self.primitive())
        return (
            [tp, "null"] if self.optional and not has_default else tp
        )  # and not has_default

    def map_cwl_type(self, parameter: cwl.Parameter) -> cwl.Parameter:
        if not NativeTypes.is_valid(self.primitive()):
            raise Exception(
                f"{self.id()} must declare its primitive as one of the NativeTypes "
                f"({', '.join(NativeTypes.all)})"
            )

        tp = NativeTypes.map_to_cwl(self.primitive())
        parameter.type = [tp, "null"] if self.optional else tp
        parameter.secondaryFiles = self.secondary_files()
        return parameter

    def cwl_input(self, value: Any):
        return value

    def wdl(self, has_default=False) -> WdlType:
        qm = self._question_mark_if_optional(has_default)
        return WdlType.parse_type(NativeTypes.map_to_wdl(self.primitive()) + qm)

    def parse_value(self, valuetoparse):
        """
        Sometimes it's useful for a value to be parsed if possible by this class.
        Eg, number, or arrays.
        """
        return valuetoparse

    # def default(self):
    #     return self.default_value



"""
Generates a CommandTool python stub from a CWL. It's not perfect, but does about 90% of the work for 90% of files.
You will need to import the data_types it uses when copying it into Python, and correct any errors.

It's not parsing requirements, so may not pick up some hints correctly.

PLEASE make sure you adjust your data_types, this only knows about primitives and WON'T HANDLE PRIMITIVES.
"""

CommandToolFormat = """
class {toolClass}(CommandTool):
{required_inputs}

{outputs}

    @staticmethod
    def tool():
        return "{tool}"

    @staticmethod
    def base_command():
        return {baseCommand}

    @staticmethod
    def container():
        return {container}

    def doc(self):
        return {doc}

    def arguments(self):
        return [{arguments}]

{optional_inputs}
"""

ToolInputFormat = '{id} = ToolInput("{id}", {typ}{args})'
ToolOutputFormat = '{id} = ToolOutput("{id}", {typ}{args})'
ToolArgumentFormat = 'ToolArgument("{value}"{args})'


def parse_cwl_dict(d):
    tool = d.get("id")

    if tool is None:
        tool = str(input("Tool name: "))

    inputs = []
    optional_inputs = []
    outputs = []
    arguments = []
    docker = d.get("dockerRequirement")
    base_command = d.get("baseCommand")
    doc = d.get("doc")
    if doc is not None:
        doc = doc.replace("\n", "")

    d_inputs = d["inputs"]
    d_outputs = d.get("outputs")
    d_arguments = d.get("arguments")

    if d_inputs is not None:
        for i in d_inputs:
            if isinstance(i, str):
                inp = d_inputs[i]
            else:
                inp = i
                i = inp.get("id")

            args = []
            if isinstance(inp, str):
                typ = inp
            else:

                typ = str(inp.get("type"))

                inpb = inp.get("inputBinding")
                if inpb is not None:
                    if "position" in inpb:
                        args.append("position=" + str(inpb["position"]))
                    if "prefix" in inpb:
                        pre = str(inpb["prefix"])
                        args.append(f'prefix="{pre}"')
                        if pre.endswith("="):
                            args.append("separate_value_from_prefix=False")

                if "doc" in inp:
                    args.append(f'doc="{inp["doc"]}"')

                if "default" in inp:
                    args.append(f'default="{inp["default"]}"')

            optional = "?" in typ or "default" in inp

            if "[]" in typ:
                opt = ", optional=True" if optional else ""
                typ = (
                    "Array("
                    + typ.replace("[]", "").replace("?", "").title()
                    + "()"
                    + opt
                    + ")"
                )
            else:
                opt = "optional=True" if optional else ""
                typ = typ.replace("?", "").title() + "(" + opt + ")"

            s = ToolInputFormat.format(
                id=i, typ=typ, args="".join((", " + x) for x in args)
            )
            if optional:
                optional_inputs.append(s)
            else:
                inputs.append(s)

    if d_outputs is not None:
        for o in d_outputs:
            if isinstance(o, str):
                out = d_outputs[o]
            else:
                out = o
                o = out.get("id")

            typ = out["type"]
            optional = "?" in typ

            if "[]" in typ:
                opt = ", optional=True" if optional else ""
                typ = (
                    "Array("
                    + typ.replace("[]", "").replace("?", "").title()
                    + "()"
                    + opt
                    + ")"
                )
            else:
                opt = "optional=True" if optional else ""
                typ = typ.replace("?", "").title() + "(" + opt + ")"

            args = []

            if "outputBinding" in out:
                if "glob" in out["outputBinding"]:
                    args.append(f"glob='{out['outputBinding']['glob']}'")

            outputs.append(
                ToolOutputFormat.format(
                    id=o, typ=typ, args="".join((", " + x) for x in args)
                )
            )

    if d_arguments is not None:
        for a in d_arguments:
            args = []
            value = a.get("valueFrom")
            if "position" in a:
                args.append("position=" + str(a["position"]))
            if "prefix" in a:
                args.append(f'prefix="{a["prefix"]}"')

            arguments.append(
                ToolArgumentFormat.format(
                    value=value, args="".join((", " + x) for x in args)
                )
            )
    print(
        CommandToolFormat.format(
            toolClass=tool.title(),
            tool=tool,
            baseCommand=base_command,
            required_inputs="\n".join(("    " + i) for i in inputs),
            optional_inputs="\n".join(("    " + i) for i in optional_inputs),
            outputs="\n".join(("    " + o) for o in outputs),
            arguments=", ".join(arguments),
            docker=docker,
            doc=f'"{doc}"',
        )
    )
    # print("\n".join(("    " + i) for i in optional_inputs))


if __name__ == "__main__":
    import argparse, yaml

    parser = argparse.ArgumentParser()
    parser.add_argument("tool")
    args = parser.parse_args()
    print(args)

    with open(args.tool) as tool:
        d = yaml.load(tool)

    parse_cwl_dict(d)



# process for adding new hints!
from abc import ABC, abstractmethod
from typing import List, Type


class Hint(ABC):
    # string inputs
    @staticmethod
    @abstractmethod
    def key():
        raise Exception("Must implement key() method")


class HintEnum(Hint, ABC):
    @staticmethod
    @abstractmethod
    def symbols():
        raise Exception("Must implement symbols() method")


class HintArray(Hint, ABC):
    @staticmethod
    @abstractmethod
    def items():
        raise Exception("Must implement items() method")


class CaptureType(HintEnum):

    TARGETED = "targeted"
    EXOME = "exome"
    CHROMOSOME = "chromosome"
    THIRTYX = "30x"
    NINETYX = "90x"
    THREEHUNDREDX = "300x"

    @staticmethod
    def key():
        return "captureType"

    @staticmethod
    def symbols():
        return [
            CaptureType.TARGETED,
            CaptureType.EXOME,
            CaptureType.CHROMOSOME,
            CaptureType.THIRTYX,
            CaptureType.NINETYX,
            CaptureType.THREEHUNDREDX,
        ]


class Engine(HintEnum):
    CROMWELL = "cromwell"

    @staticmethod
    def key():
        return "engine"

    @staticmethod
    def symbols():
        return [Engine.CROMWELL]


HINTS: List[Type[Hint]] = [CaptureType, Engine]



from typing import Dict

from janis_core.__meta__ import GITHUB_URL
from janis_core.translations.exportpath import ExportPathKeywords
from janis_core.enums.supportedtranslations import (
    SupportedTranslations,
    SupportedTranslation,
)
from janis_core.translations.wdl import WdlTranslator
from .cwl import CwlTranslator
from .translationbase import TranslatorBase


def get_translator(translation: SupportedTranslation) -> TranslatorBase:
    if translation == SupportedTranslations.CWL:
        return CwlTranslator()
    elif translation == SupportedTranslations.WDL:
        return WdlTranslator()

    raise NotImplementedError(
        f"The requested translation ('{translation}') has not been implemented yet, "
        f"why not contribute one at '{GITHUB_URL}'."
    )


def translate_workflow(
    workflow,
    translation: SupportedTranslation,
    to_console=True,
    tool_to_console=False,
    with_docker=True,
    with_resource_overrides=False,
    to_disk=False,
    write_inputs_file=True,
    export_path=ExportPathKeywords.default,
    should_validate=False,
    should_zip=True,
    merge_resources=False,
    hints=None,
    allow_null_if_not_optional=True,
    additional_inputs: Dict = None,
    max_cores=None,
    max_mem=None,
):
    translator = get_translator(translation)
    return translator.translate(
        workflow,
        to_console=to_console,
        tool_to_console=tool_to_console,
        with_docker=with_docker,
        with_resource_overrides=with_resource_overrides,
        to_disk=to_disk,
        export_path=export_path,
        write_inputs_file=write_inputs_file,
        should_validate=should_validate,
        should_zip=should_zip,
        merge_resources=merge_resources,
        hints=hints,
        allow_null_if_not_optional=allow_null_if_not_optional,
        additional_inputs=additional_inputs,
        max_cores=max_cores,
        max_mem=max_mem,
    )


def translate_tool(
    tool,
    translation: SupportedTranslation,
    to_console=True,
    to_disk=False,
    export_path=None,
    with_docker=True,
    with_resource_overrides=False,
    max_cores=None,
    max_mem=None,
) -> str:
    translator = get_translator(translation)
    return translator.translate_tool(
        tool,
        to_console=to_console,
        to_disk=to_disk,
        export_path=export_path,
        with_docker=with_docker,
        with_resource_overrides=with_resource_overrides,
        max_cores=max_cores,
        max_mem=max_mem,
    )


def build_resources_input(
    workflow, translation: SupportedTranslation, hints, max_cores=None, max_mem=None
) -> str:
    translator = get_translator(translation)
    return translator.stringify_translated_inputs(
        translator.build_resources_input(
            workflow, hints, max_cores=max_cores, max_mem=max_mem
        )
    )



"""
Welcome to the source code for Janis!

Janis is a framework creating specialised, simple workflow definitions that are
transpiled to Common Workflow Language or Workflow Definition Language.

Below you'll find the classes you can access by importing Janis:

>>  import janis as j

Some noteworthy classes are Workflow, CommandTool and the janis.translations module

Some terminology:
    - Edge:
        - An edge may exist between 2 nodes, it represents a dependency
        - Every edge will have a source_map which is indexed by the tag of the input on the finish node
        - This source_map value is singular or a list, as you can connect multiple sources to one input



_Proudly made on Planet Earth._

"""

import sys
import pkg_resources


# PEP396:  https://www.python.org/dev/peps/pep-0396/
from janis_core.__meta__ import __version__
from janis_core.registry.shed import JanisShed
import janis_core.registry.entrypoints as entrypoints
from janis_core.utils.scatter import ScatterDescription, ScatterMethods
from janis_core.hints import CaptureType, Engine, HINTS, Hint, HintEnum, HintArray
from janis_core.tool.commandtool import CommandTool, ToolArgument, ToolInput, ToolOutput
from janis_core.tool.tool import Tool, ToolTypes
from janis_core.translations import SupportedTranslations
from janis_core.types import (
    InputSelector,
    WildcardSelector,
    MemorySelector,
    CpuSelector,
    StringFormatter,
)
from janis_core.types.common_data_types import (
    Boolean,
    String,
    Int,
    Float,
    Double,
    File,
    Directory,
    Array,
    Filename,
    Stdout,
    Stderr,
)
from janis_core.types.data_types import DataType
from janis_core.utils import get_value_for_hints_and_ordered_resource_tuple
from janis_core.utils.logger import Logger, LogLevel
from janis_core.utils.metadata import Metadata, WorkflowMetadata, ToolMetadata

from janis_core.workflow.workflow import Workflow, WorkflowBuilder

# from janis_core.workflow.input import Input
# from janis_core.workflow.output import Output
# from janis_core.workflow.step import Step
# from janis_core.workflow.workflow import Workflow



###################
# Implementations #
###################
from inspect import isclass
from typing import Dict, Any, Union, Type, List

import wdlgen
import cwlgen

from janis_core.types.data_types import (
    DataType,
    NativeTypes,
    NativeType,
    PythonPrimitive,
)


ParseableTypeBase = Union[Type[PythonPrimitive], DataType, Type[DataType]]
ParseableType = ParseableTypeBase


class String(DataType):
    @staticmethod
    def name():
        return "String"

    @staticmethod
    def primitive():
        return NativeTypes.kStr

    @staticmethod
    def doc():
        return "A string"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "string", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))

    def can_receive_from(self, other, source_has_default=False):
        if isinstance(other, Filename):
            return True
        return super().can_receive_from(other, source_has_default=source_has_default)

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool):
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, str)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None

        return f"Value was of type {type(meta)}, expected string"


class Filename(String):
    def __init__(
        self, prefix="generated", suffix=None, extension: str = None, guid: str = None
    ):
        """
        :param suffix: suffix the guid
        :param extension: with no '.' (dot)
        :param guid: Use this guid instead of generating one
        """
        import uuid

        self.prefix = prefix
        self.extension = extension
        self.suffix = suffix
        self.guid = guid if guid is not None else str(uuid.uuid1())
        super().__init__(optional=True)

    @staticmethod
    def name() -> str:
        return "Filename"

    @staticmethod
    def primitive() -> NativeType:
        return NativeTypes.kStr

    def cwl_type(self, has_default=False):
        self.optional = False
        t = super().cwl_type()
        self.optional = True
        return t

    @staticmethod
    def doc() -> str:
        return """
This class is a placeholder for generated filenames, by default it is optional and CAN be overrided, 
however the program has been structured in a way such that these names will be generated based on the step label. 
These should only be used when the tool _requires_ a filename to output and you aren't 
concerned what the filename should be. The Filename DataType should NOT be used as an output.
""".strip()

    @classmethod
    def schema(cls) -> Dict:
        pass

    def map_cwl_type(self, parameter: cwlgen.Parameter):
        super().map_cwl_type(parameter)
        parameter.default = self.generated_filenamecwl()

    def generated_filename(self) -> str:

        pre = self.prefix
        guid = self.guid
        suf = self.suffix if self.suffix else ""
        ex = "" if self.extension is None else self.extension

        dash1 = "-" if self.prefix else ""

        return pre + dash1 + guid + suf + ex

    def generated_filenamecwl(self) -> str:
        code = "Math.random().toString(16).substring(2, 8)"
        pf = (self.prefix + "-") if self.prefix else ""
        sf = self.suffix if self.suffix else ""
        ext = self.extension if self.extension else ""
        return f'"{pf}generated-" + {code} + "{sf + ext}"'

    def can_receive_from(self, other: DataType, source_has_default=False):
        # Specific override because Filename should be able to receive from string
        if isinstance(other, String):
            return True  # Always provides default, and is always optional
        return super().can_receive_from(other, source_has_default=source_has_default)

    def wdl(self, has_default=True):
        return wdlgen.WdlType.parse_type(NativeTypes.map_to_wdl(self.primitive()))

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool):
        return True

    def invalid_value_hint(self, meta):
        return None


class Int(DataType):
    @staticmethod
    def name():
        return "Integer"

    @staticmethod
    def primitive():
        return NativeTypes.kInt

    def doc(self):
        return "An integer"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "number", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool):
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, int)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None
        return f"Value was of type {type(meta)}, expected int"


class Float(DataType):
    @staticmethod
    def name():
        return "Float"

    @staticmethod
    def primitive():
        return NativeTypes.kFloat

    def doc(self):
        return "A float"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "number", "required": True}

    def input_field_from_input(self, meta: Dict):
        return next(iter(meta.values()))

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, float) or isinstance(meta, int)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None
        return f"Value was of type {type(meta)}, expected float | int"


class Double(DataType):
    @staticmethod
    def name():
        return "Double"

    @staticmethod
    def primitive():
        return NativeTypes.kDouble

    def doc(self):
        return "An integer"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "number", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, float) or isinstance(meta, int)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None
        return f"Value was of type {type(meta)}, expected float | int"


class Boolean(DataType):
    @staticmethod
    def name():
        return "Boolean"

    @staticmethod
    def primitive():
        return NativeTypes.kBool

    def doc(self):
        return "A boolean"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "boolean", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, bool)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None
        return f"Value was of type {type(meta)}, expected bool"


class File(DataType):
    def __init__(self, optional=False, extension=None):
        """
        :param optional:
        :param common_extension: Used in CWL to try and guess the file extension where it's not available otherwise
        """
        super(File, self).__init__(optional=optional)
        self.extension = extension

    @staticmethod
    def name():
        return "File"

    @staticmethod
    def primitive():
        return NativeTypes.kFile

    def doc(self):
        return "A local file"

    @classmethod
    def schema(cls) -> Dict:
        return {"path": {"type": "string", "required": True}}

    def get_value_from_meta(self, meta):
        return meta.get("path")

    def cwl_input(self, value: Any):
        return {"class": cwlgen.CwlTypes.FILE, "path": value}

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, str)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None

        return f"Value was of type {type(meta)}, expected string (path)"


class Directory(DataType):
    def __init__(self, optional=False):
        """
        Specifically exclude default
        """
        super(Directory, self).__init__(optional=optional)

    @staticmethod
    def name():
        return "Directory"

    @staticmethod
    def primitive():
        return NativeTypes.kDirectory

    def doc(self):
        return "A directory of files"

    def get_value_from_meta(self, meta):
        return meta["path"]

    @classmethod
    def schema(cls) -> Dict:
        return {"path": {"type": "string", "required": True}}

    def input_field_from_input(self, meta):
        return meta["path"]

    def cwl_input(self, value: Any):
        # WDL: "{workflowName}.label" = meta["path"}
        return {"class": cwlgen.CwlTypes.DIRECTORY, "path": value}

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        if meta is None:
            return self.optional or allow_null_if_not_optional
        return isinstance(meta, str)

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None
        return f"Value was of type {type(meta)}, expected string (path)"


class Array(DataType):
    def __init__(self, t: ParseableType, optional=False):
        resolvedtype = get_instantiated_type(t)
        if not isinstance(resolvedtype, DataType):
            raise Exception(f"Type t ({type(t)}) must be an instance of 'DataType'")

        self._t = resolvedtype
        super().__init__(optional)

    def subtype(self):
        return self._t

    @staticmethod
    def name():
        return "Array"

    @staticmethod
    def primitive():
        return NativeTypes.kArray

    def id(self):
        if self._t is None:
            return super().id()
        t = self._t
        typed = f"Array<{t.id()}>"
        if self.optional:
            return f"Optional<{typed}>"
        return typed

    def doc(self):
        return "An array"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "array"}

    def cwl_type(self, has_default=False):
        inp = cwlgen.CommandInputArraySchema(
            items=self._t.cwl_type(),
            # label=None,
            # input_binding=None
        )
        return [inp, "null"] if self.optional and not has_default else inp

    def map_cwl_type(self, parameter: cwlgen.Parameter) -> cwlgen.Parameter:
        parameter.type = cwlgen.CommandInputArraySchema(
            items=None, label=None, input_binding=None
        )
        return parameter

    def cwl_input(self, value: Any):
        if isinstance(value, list):
            return [self._t.cwl_input(v) for v in value]
        if value is None:
            return None
        else:
            raise Exception(f"Input value for input '{self.id()}' was not an array")

    def wdl(self, has_default=False) -> wdlgen.WdlType:
        ar = wdlgen.ArrayType(self._t.wdl(has_default=False), requires_multiple=False)
        return wdlgen.WdlType(ar, optional=self.optional or has_default)

    def can_receive_from(self, other, source_has_default=False):
        if isinstance(other, Array):
            return self._t.can_receive_from(other._t)
        if not self._t.can_receive_from(other):
            return False
        return super().can_receive_from(other, source_has_default=source_has_default)

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        if meta is None:
            return self.optional or allow_null_if_not_optional
        if not isinstance(meta, list):
            return False
        return all(
            self.subtype().validate_value(q, allow_null_if_not_optional) for q in meta
        )

    def invalid_value_hint(self, meta):
        if meta is None:
            return "value was null"
        if self.validate_value(meta, True):
            return None

        if not isinstance(meta, list):
            return f"Value was of type {type(meta)}, expected type Array<{self.subtype().id()}>"

        hints = []
        st = self.subtype()
        for i in range(len(meta)):
            hint = st.invalid_value_hint(meta[i])
            if not hint:
                continue
            hints.append(f"{i}. {hint}")

        return str(hints)

    def parse_value(self, valuetoparse):
        if not isinstance(valuetoparse, list):
            valuetoparse = [valuetoparse]

        return [self.subtype().parse_value(v) for v in valuetoparse]


class Stdout(File):
    @staticmethod
    def name():
        return "Stdout"

    def __init__(self, subtype=None, stdoutname=None):
        super().__init__(optional=False)

        subtype = get_instantiated_type(subtype) if subtype is not None else File()

        if subtype and not isinstance(subtype, File):
            raise Exception(
                "Janis does not currently support non-File stdout annotations"
            )

        self.stdoutname = stdoutname
        self.subtype = subtype

        if self.subtype.secondary_files():
            raise Exception(
                f"The subtype '{self.subtype.__name__}' has secondary files, "
                f"but stdout does not have the ability to collect files"
            )

    @staticmethod
    def primitive():
        return NativeTypes.kStdout

    def id(self):
        return f"stdout<{self.subtype.id()}>"

    def received_type(self):
        return self.subtype

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        """
        Will always toss away the value
        """
        return True

    def invalid_value_hint(self, meta):
        return None


class Stderr(File):
    @staticmethod
    def name():
        return "Stderr"

    def __init__(self, subtype=None, stderrname=None):
        super().__init__(optional=False)

        subtype = get_instantiated_type(subtype) if subtype is not None else File()

        if subtype and not isinstance(subtype, File):
            raise Exception(
                "Janis does not currently support non-File stderr annotations"
            )

        self.stderrname = stderrname
        self.subtype = subtype

        if self.subtype.secondary_files():
            raise Exception(
                f"The subtype '{self.subtype.__name__}' has secondary files, "
                f"but stderr does not have the ability to collect files"
            )

    @staticmethod
    def primitive():
        return NativeTypes.kStderr

    def id(self):
        return f"stderr<{self.subtype.id()}>"

    def received_type(self):
        return self.subtype

    def validate_value(self, meta: Any, allow_null_if_not_optional: bool) -> bool:
        """
        Will always toss away the value
        """
        return True

    def invalid_value_hint(self, meta):
        return None


all_types = [
    String,
    Filename,
    Int,
    Float,
    Double,
    Boolean,
    File,
    Directory,
    Stdout,
    Array,
]


def get_instantiated_type(datatype: ParseableType):

    if isinstance(datatype, list):
        if len(datatype) == 0:
            raise TypeError("Couldn't determine type of array with length 0")
        return Array(get_instantiated_type(datatype[0]))

    if isinstance(datatype, DataType):
        return datatype

    if isclass(datatype) and issubclass(datatype, DataType):
        return datatype()

    typedt = type(datatype)
    if datatype == str or typedt == str:
        return String()
    if datatype == bool or typedt == bool:
        return Boolean()
    if datatype == int or typedt == int:
        return Int()
    if datatype == float or typedt == float:
        return Float()

    raise TypeError(f"Unable to parse type '{str(datatype)}'")



"""
    Logger - Controls logging of the application
"""
import sys
import os
from datetime import datetime
from typing import Optional, TextIO


class _bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class LogLevel:
    CRITICAL = 2  # RED
    WARNING = 3  # YELLOW
    INFO = 4  # WHITE
    DEBUG = 5  # GREY

    @staticmethod
    def get_color(level: int):
        if level == LogLevel.CRITICAL:
            return _bcolors.FAIL

        if level == LogLevel.WARNING:
            return _bcolors.WARNING

        if level == LogLevel.INFO:
            return _bcolors.OKBLUE

        if level == LogLevel.DEBUG:
            return _bcolors.ENDC

        return _bcolors.ENDC

    @staticmethod
    def get_str(level: int):
        if level == LogLevel.CRITICAL:
            return "CRITICAL"

        if level == LogLevel.WARNING:
            return "WARN"

        if level == LogLevel.INFO:
            return "INFO"

        if level == LogLevel.DEBUG:
            return "DEBUG"

        return ""

    @staticmethod
    def from_str(level: str):
        level = level.upper()
        if level == "CRITICAL":
            return LogLevel.CRITICAL
        if level == "WARN":
            return LogLevel.WARNING
        if level == "INFO":
            return LogLevel.INFO
        if level == "DEBUG":
            return LogLevel.DEBUG
        if level == "NONE":
            return None

        raise KeyError(
            f"Couldn't detect LogLevel type '{level}', expected one of: "
            f"CRITICAL, WARN, INFO, DEBUG, NONE"
        )


class Logger:
    CONSOLE_LEVEL: Optional[int] = LogLevel.INFO
    __TEMP_CONSOLE_LEVEL: Optional[int] = None
    WRITE_LEVEL: Optional[int] = LogLevel.DEBUG

    WRITE_LOCATION: Optional[str] = None
    __WRITE_POINTER: Optional[TextIO] = None

    @staticmethod
    def set_console_level(level: Optional[int]):
        Logger.CONSOLE_LEVEL = level

    @staticmethod
    def mute():
        """
        Decreases log level to critical until unmute is called
        """
        Logger.__TEMP_CONSOLE_LEVEL = Logger.CONSOLE_LEVEL
        Logger.set_console_level(LogLevel.CRITICAL)

    @staticmethod
    def unmute():
        """
        unmute the console, if __TEMP is None, should not do anything
        :return:
        """
        if Logger.__TEMP_CONSOLE_LEVEL is not None:
            Logger.set_console_level(Logger.__TEMP_CONSOLE_LEVEL)

    @staticmethod
    def set_write_level(level: Optional[int]):
        Logger.WRITE_LEVEL = level

    @staticmethod
    def set_write_location(location: str):
        if Logger.__WRITE_POINTER is not None:
            Logger.__WRITE_POINTER.close()

        if location is None:
            Logger.WRITE_LOCATION = None
            Logger.__WRITE_POINTER = None
            return

        Logger.WRITE_LOCATION = location
        Logger.__WRITE_POINTER = open(location, "a")

    @staticmethod
    def close_file():
        if Logger.__WRITE_POINTER is not None:
            Logger.__WRITE_POINTER.close()
            Logger.__WRITE_POINTER = None

    @staticmethod
    def log(message: str, level: int = LogLevel.DEBUG):
        if level is None:
            # This is a developer error, we should never try to log with no level, it's purely for
            return

        m = f"{Logger.get_prefix(level)}: {message}"
        if Logger.CONSOLE_LEVEL is not None and level <= Logger.CONSOLE_LEVEL:
            print(LogLevel.get_color(level) + m + _bcolors.ENDC, file=sys.stderr)

        if (
            Logger.WRITE_LEVEL is not None
            and level <= Logger.WRITE_LEVEL
            and Logger.__WRITE_POINTER is not None
            and not Logger.__WRITE_POINTER.closed
        ):
            Logger.__WRITE_POINTER.write(m + "\n")
            Logger.__WRITE_POINTER.flush()
            os.fsync(Logger.__WRITE_POINTER.fileno())

    @staticmethod
    def info(message: str):
        Logger.log(message, LogLevel.INFO)

    @staticmethod
    def warn(message: str):
        Logger.log(message, LogLevel.WARNING)

    @staticmethod
    def critical(message: str):
        Logger.log(message, LogLevel.CRITICAL)

    @staticmethod
    def log_ex(ex: Exception):
        Logger.critical(str(ex))

    @staticmethod
    def get_prefix(level: int):
        return f"{datetime.now().replace(microsecond=0).isoformat()} [{LogLevel.get_str(level)}]"



"""
WDL

This is one of the more complicated classes, it takes the janis in-memory representation of a workflow,
and converts it into the equivalent WDL objects. Python-wdlgen distances us from the memory representation
and the actual string-specification.

This file is logically structured similar to the cwl equiv:

- Imports
- dump_wdl
- translate_workflow
- translate_tool (command tool)
- other translate methods
- selector helpers (InputSelector, WildcardSelector, CpuSelector, MemorySelector)
- helper methods
"""

import json
from inspect import isclass
from uuid import uuid4
from typing import List, Dict, Optional, Any, Set, Tuple

import wdlgen as wdl
from janis_core.utils.scatter import ScatterDescription, ScatterMethod, ScatterMethods

from janis_core.graph.steptaginput import Edge, StepTagInput
from janis_core.tool.commandtool import CommandTool, ToolInput, ToolArgument, ToolOutput
from janis_core.tool.tool import Tool, TOutput
from janis_core.translations.translationbase import TranslatorBase
from janis_core.types import (
    InputSelector,
    WildcardSelector,
    CpuSelector,
    MemorySelector,
    StringFormatter,
    String,
)
from janis_core.types.common_data_types import (
    Stdout,
    Stderr,
    Array,
    Boolean,
    Filename,
    File,
    Int,
)
from janis_core.utils import first_value, recursive_2param_wrap
from janis_core.utils.generators import generate_new_id_from
from janis_core.utils.logger import Logger
from janis_core.utils.validators import Validators

# from janis_core.workflow.step import StepNode


## PRIMARY TRANSLATION METHODS


class WdlTranslator(TranslatorBase):
    def __init__(self):
        super().__init__(name="wdl")

    @staticmethod
    def stringify_translated_workflow(wf):
        return wf.get_string()

    @staticmethod
    def stringify_translated_tool(tool):
        return tool.get_string()

    @staticmethod
    def stringify_translated_inputs(inputs):
        return json.dumps(inputs, sort_keys=True, indent=4, separators=(",", ": "))

    @staticmethod
    def validate_command_for(wfpath, inppath, tools_dir_path, tools_zip_path):
        return ["java", "-jar", "$womtooljar", "validate", wfpath]

    @classmethod
    def translate_workflow(
        cls, wfi, with_docker=True, with_resource_overrides=False, is_nested_tool=False
    ) -> Tuple[any, Dict[str, any]]:
        """
        Translate the workflow into wdlgen classes!


        :param with_resource_overrides:
        :param with_docker:
        :param is_nested_tool:
        :return:
        """
        # Import needs to be here, otherwise we end up circularly importing everything
        # I need the workflow for type comparison
        from janis_core.workflow.workflow import Workflow

        wf: Workflow = wfi

        # Notes:
        #       All wdlgen classes have a .get_string(**kwargs) function
        #       The wdlgen Workflow class requires a

        # As of 2019-04-16: we use development (Biscayne) features
        # like Directories and wdlgen uses the new input {} syntax
        w = wdl.Workflow(wf.id(), version="development")
        tools: List[Tool] = [s.tool for s in wf.step_nodes.values()]

        inputs = list(wf.input_nodes.values())
        steps = list(wf.step_nodes.values())
        outputs = list(wf.output_nodes.values())

        wtools = {}  # Store all the tools by their name in this dictionary
        tool_aliases, step_aliases = build_aliases(
            wf.step_nodes.values()
        )  # Generate call and import aliases

        # Convert self._inputs -> wdl.Input
        for i in inputs:
            dt = i.datatype
            expr = None
            if isinstance(i.datatype, Filename):
                # expr = f'"{i.datatype.generated_filename()}"'
                dt = String(optional=True)

            wd = dt.wdl(has_default=i.default is not None)

            w.inputs.append(
                wdl.Input(
                    data_type=wd, name=i.id(), expression=expr, requires_quotes=False
                )
            )

            is_array = isinstance(i.datatype, Array)
            if i.datatype.secondary_files() or (
                is_array and i.datatype.subtype().secondary_files()
            ):
                secs = (
                    i.datatype.secondary_files()
                    if not is_array
                    else i.datatype.subtype().secondary_files()
                )

                w.inputs.extend(
                    wdl.Input(wd, get_secondary_tag_from_original_tag(i.id(), s))
                    for s in secs
                )

        resource_inputs = []
        if with_resource_overrides:
            resource_inputs = build_resource_override_maps_for_workflow(wf)
            w.inputs.extend(resource_inputs)

        # Convert self._outputs -> wdl.Output
        for o in outputs:
            sourcenode, sourcetag = o.source
            w.outputs.append(
                wdl.Output(
                    o.datatype.wdl(),
                    o.id(),
                    "{a}.{b}".format(  # Generate fully qualified stepid.tag identifier (MUST be step node)
                        a=sourcenode.id(), b=sourcetag
                    ),
                )
            )
            if o.datatype.secondary_files():
                w.outputs.extend(
                    wdl.Output(
                        o.datatype.wdl(),
                        get_secondary_tag_from_original_tag(o.id(), s),
                        "{a}.{b}".format(  # Generate fully qualified stepid.tag identifier (MUST be step node)
                            a=sourcenode.id(),
                            b=get_secondary_tag_from_original_tag(sourcetag, s),
                        ),
                    )
                    for s in o.datatype.secondary_files()
                )

        # Generate import statements (relative tool dir is later?)
        uniquetoolmap: Dict[str, Tool] = {t.id(): t for t in tools}
        w.imports = [
            wdl.Workflow.WorkflowImport(
                t.id(),
                tool_aliases[t.id().lower()].upper(),
                None if is_nested_tool else "tools/",
            )
            for t in uniquetoolmap.values()
        ]

        # Step[] -> (wdl.Task | wdl.Workflow)[]
        forbiddenidentifiers = set(
            [i.id() for i in inputs]
            + list(tool_aliases.keys())
            + list(s.id() for s in steps)
        )
        for s in steps:
            t = s.tool

            if t.id() not in wtools:
                if isinstance(t, Workflow):
                    wf_wdl, wf_tools = cls.translate_workflow(
                        t,
                        with_docker=with_docker,
                        is_nested_tool=True,
                        with_resource_overrides=with_resource_overrides,
                    )
                    wtools[t.id()] = wf_wdl
                    wtools.update(wf_tools)

                elif isinstance(t, CommandTool):
                    wtools[t.id()] = cls.translate_tool_internal(
                        t,
                        with_docker=with_docker,
                        with_resource_overrides=with_resource_overrides,
                    )

            resource_overrides = {}

            if with_resource_overrides:
                toolroverrides = build_resource_override_maps_for_tool(t)
                for r in toolroverrides:
                    resource_overrides[r.name] = s.id() + "_" + r.name

            call = translate_step_node(
                s,
                tool_aliases[t.id().lower()].upper() + "." + t.id(),
                resource_overrides,
                forbiddenidentifiers,
            )

            w.calls.append(call)

        return w, wtools

    @classmethod
    def translate_tool_inputs(cls, toolinputs: List[ToolInput]):
        ins = []
        for i in toolinputs:
            wd = i.input_type.wdl(has_default=i.default is not None)
            expr = None
            if isinstance(i.input_type, Filename):
                expr = f'"{i.input_type.generated_filename()}"'
            if isinstance(wd, list):
                ins.extend(wdl.Input(w, i.id()) for w in wd)
            else:

                ins.append(wdl.Input(wd, i.id(), expr, requires_quotes=False))

                sec = value_or_default(
                    i.input_type.subtype().secondary_files()
                    if isinstance(i.input_type, Array)
                    else i.input_type.secondary_files(),
                    default=[],
                )
                ins.extend(
                    wdl.Input(wd, get_secondary_tag_from_original_tag(i.id(), s))
                    for s in sec
                )
        return ins

    @classmethod
    def translate_tool_outputs(
        cls, tooloutputs: List[ToolOutput], inputsmap: Dict[str, ToolInput], toolid
    ):
        outs: List[wdl.Output] = []

        for o in tooloutputs:
            outs.extend(
                translate_output_node_with_glob(o, o.glob, inputsmap, toolId=toolid)
            )
        return outs

    @classmethod
    def translate_tool_args(
        cls, toolargs: List[ToolArgument], inpmap: Dict[str, ToolInput], **debugkwargs
    ):
        if not toolargs:
            return []
        commandargs = []
        for a in toolargs:
            val = get_input_value_from_potential_selector_or_generator(a.value, inpmap)
            should_wrap_in_quotes = isinstance(val, str) and (
                a.shell_quote is None or a.shell_quote
            )
            wrapped_val = f"'{val}'" if should_wrap_in_quotes else val
            commandargs.append(
                wdl.Task.Command.CommandArgument(a.prefix, wrapped_val, a.position)
            )
        return commandargs

    @classmethod
    def build_command_from_inputs(
        cls, toolinputs: List[ToolInput], inputmap: Dict[str, ToolInput]
    ):
        command_ins = []
        for i in toolinputs:
            cmd = translate_command_input(i, inputmap)
            if cmd:
                command_ins.append(cmd)
        return command_ins

    @classmethod
    def translate_tool_internal(
        cls, tool: CommandTool, with_docker=True, with_resource_overrides=False
    ):

        if not Validators.validate_identifier(tool.id()):
            raise Exception(
                f"The identifier '{tool.id()}' for class '{tool.__class__.__name__}' was not validated by "
                f"'{Validators.identifier_regex}' (must start with letters, and then only contain letters, "
                f"numbers or an underscore)"
            )

        inputs: List[ToolInput] = [*cls.get_resource_override_inputs(), *tool.inputs()]
        inmap = {i.tag: i for i in inputs}

        ins: List[wdl.Input] = cls.translate_tool_inputs(inputs)
        outs: List[wdl.Output] = cls.translate_tool_outputs(
            tool.outputs(), inmap, tool.id()
        )
        command_args = cls.translate_tool_args(
            tool.arguments(), inmap, toolId=tool.id()
        )
        command_ins = cls.build_command_from_inputs(tool.inputs(), inmap)

        commands = []

        env = tool.env_vars()
        if env:
            commands.extend(
                prepare_env_var_setters(env, inputsdict=inmap, toolid=tool.id())
            )

        for ti in tool.inputs():
            if ti.localise_file:
                commands.extend(prepare_move_statement_for_input_to_localise(ti))
            else:
                commands.extend(prepare_move_statement_for_potential_secondaries(ti))

        rbc = tool.base_command()
        bc = " ".join(rbc) if isinstance(rbc, list) else rbc

        commands.append(wdl.Task.Command(bc, command_ins, command_args))

        r = wdl.Task.Runtime()
        if with_docker:
            r.add_docker(tool.container())

        # These runtime kwargs cannot be optional, but we've enforced non-optionality when we create them
        r.kwargs["cpu"] = get_input_value_from_potential_selector_or_generator(
            CpuSelector(), inmap, string_environment=False, id="runtimestats"
        )
        r.kwargs["memory"] = wdl.IfThenElse(
            "defined(runtime_memory)", '"${runtime_memory}G"', '"4G"'
        )

        if with_resource_overrides:
            ins.append(wdl.Input(wdl.WdlType.parse_type("String"), "runtime_disks"))
            r.kwargs["disks"] = "runtime_disks"
            r.kwargs["zones"] = '"australia-southeast1-b"'

        r.kwargs["preemptible"] = 2

        return wdl.Task(tool.id(), ins, outs, commands, r, version="development")

    @classmethod
    def build_inputs_file(
        cls,
        workflow,
        recursive=False,
        merge_resources=False,
        hints=None,
        additional_inputs: Dict = None,
        max_cores=None,
        max_mem=None,
    ) -> Dict[str, any]:
        """
        Recursive is currently unused, but eventually input overrides could be generated the whole way down
        a call chain, including subworkflows: https://github.com/openwdl/wdl/issues/217
        :param merge_resources:
        :param recursive:
        :param workflow:
        :return:
        """
        inp = {}
        ad = additional_inputs or {}

        for i in workflow.input_nodes.values():
            inp_key = f"{workflow.id()}.{i.id()}"
            value = ad.get(i.id()) or i.value or i.default
            if cls.inp_can_be_skipped(i, value):
                continue

            inp_val = value

            inp[inp_key] = inp_val
            if i.datatype.secondary_files():
                for sec in i.datatype.secondary_files():
                    inp[
                        get_secondary_tag_from_original_tag(inp_key, sec)
                    ] = apply_secondary_file_format_to_filename(inp_val, sec)
            elif (
                isinstance(i.datatype, Array) and i.datatype.subtype().secondary_files()
            ):
                # handle array of secondary files
                for sec in i.datatype.subtype().secondary_files():
                    inp[get_secondary_tag_from_original_tag(inp_key, sec)] = (
                        [
                            apply_secondary_file_format_to_filename(iinp_val, sec)
                            for iinp_val in inp_val
                        ]
                        if inp_val
                        else None
                    )

        if merge_resources:
            inp.update(
                cls.build_resources_input(
                    workflow, hints, max_cores, max_mem, inputs=ad
                )
            )

        return inp

    @classmethod
    def build_resources_input(
        cls, workflow, hints, max_cores=None, max_mem=None, inputs=None, prefix=None
    ):
        return super().build_resources_input(
            workflow=workflow,
            hints=hints,
            max_cores=max_cores,
            max_mem=max_mem,
            prefix=prefix or f"{workflow.id()}.",
            inputs=inputs,
        )

    @staticmethod
    def workflow_filename(workflow):
        return workflow.id() + ".wdl"

    @staticmethod
    def inputs_filename(workflow):
        return workflow.id() + "-inp.json"

    @staticmethod
    def tool_filename(tool):
        return (tool.id() if isinstance(tool, Tool) else str(tool)) + ".wdl"

    @staticmethod
    def resources_filename(workflow):
        return workflow.id() + "-resources.json"


def resolve_tool_input_value(tool_input: ToolInput, **debugkwargs):
    name = tool_input.id()
    indefault = (
        tool_input.input_type
        if isinstance(tool_input.input_type, Filename)
        else tool_input.default
    )

    default = None
    if isinstance(indefault, CpuSelector):
        if indefault.default:
            default = "if defined(runtime_cpu) then runtime_cpu else " + str(
                indefault.default
            )
        else:
            default = "runtime_cpu"

    elif isinstance(indefault, InputSelector):
        Logger.critical(
            f"WDL does not support command line level defaults that select a different input, this will remove the "
            f"value: '{indefault}' for tool_input '{tool_input.tag}'"
        )

    else:
        default = get_input_value_from_potential_selector_or_generator(
            indefault, None, string_environment=False, **debugkwargs
        )

    if default:
        name = f"if defined({name}) then {name} else {default}"

    if tool_input.localise_file:
        if isinstance(tool_input.input_type, Array):
            raise Exception(
                "Localising files through `basename(x)` is unavailable for arrays of files: https://github.com/openwdl/wdl/issues/333"
            )
        name = "basename(%s)" % name

    return name


def translate_command_input(tool_input: ToolInput, inputsdict, **debugkwargs):
    # make sure it has some essence of a command line binding, else we'll skip it
    # TODO: make a property on ToolInput (.bind_to_commandline) and set default to true
    if not (tool_input.position is not None or tool_input.prefix):
        return None

    name = resolve_tool_input_value(tool_input, **debugkwargs)
    optional = tool_input.input_type.optional or isinstance(
        tool_input.default, CpuSelector
    )
    position = tool_input.position
    separate_value_from_prefix = tool_input.separate_value_from_prefix
    prefix = tool_input.prefix
    true = None
    sep = tool_input.separator

    is_array = isinstance(tool_input.input_type, Array)
    separate_arrays = is_array and tool_input.prefix_applies_to_all_elements

    if isinstance(tool_input.input_type, Boolean):
        true = tool_input.prefix
        prefix = None

    return wdl.Task.Command.CommandInput(
        name=name,
        optional=optional,
        prefix=prefix,
        position=position,
        separate_value_from_prefix=(
            separate_value_from_prefix
            if separate_value_from_prefix is not None
            else True
        ),
        # Instead of using default, we'll use the ${if defined($var) then val1 else val2}
        # as it progress through the rest properly
        # default=default,
        true=true,
        separator=(
            None
            if not is_array or separate_arrays
            else (sep if sep is not None else " ")
        ),
        separate_arrays=separate_arrays,
    )


def translate_output_node_with_glob(
    o, glob, inputmap: Dict[str, ToolInput], **debugkwargs
) -> List[wdl.Output]:
    if isinstance(o.output_type, Stdout):
        base_expression = "stdout()"
        return [wdl.Output(o.output_type.wdl(), o.id(), base_expression)]

    if isinstance(o.output_type, Stderr):
        base_expression = "stderr()"
        return [wdl.Output(o.output_type.wdl(), o.id(), base_expression)]

    elif isinstance(glob, InputSelector):
        return translate_input_selector_for_output(o, glob, inputmap, **debugkwargs)

    elif isinstance(glob, StringFormatter):
        return translate_string_formatter_for_output(o, glob, inputmap, **debugkwargs)

    elif isinstance(glob, WildcardSelector):
        base_expression = translate_wildcard_selector(glob)
        if not isinstance(o.output_type, Array):
            Logger.warn(
                f"The command tool ({debugkwargs}).{o.tag}' used a star-bind (*) glob to find the output, "
                f"but the return type was not an array. For WDL, the first element will be used, "
                f"ie: '{base_expression}[0]'"
            )
            base_expression += "[0]"
        wdl_type = wdl.WdlType.parse_type(o.output_type.wdl())
        outputs = [wdl.Output(wdl_type, o.id(), base_expression)]

        secondary = o.output_type.secondary_files()
        if secondary:
            outputs.extend(
                wdl.Output(
                    wdl_type,
                    get_secondary_tag_from_original_tag(o.id(), s),
                    base_expression,
                )
                for s in o.output_type.secondary_files()
            )
        return outputs

    else:
        raise Exception(
            f"Tool ({debugkwargs}) has the non-selector glob: '{glob}', this is deprecated. "
            f"Please use the WildcardSelector to build output for '{o.id()}'"
        )


def translate_input_selector_for_output(
    out, selector: InputSelector, inp_map: Dict[str, ToolInput], **debugkwargs
) -> List[wdl.Output]:
    base_expression = translate_input_selector(
        selector, inp_map, string_environment=False, **debugkwargs
    )

    tool_in = inp_map.get(selector.input_to_select)
    if not tool_in:
        raise Exception(
            f"The InputSelector for tool '{debugkwargs}.{out.id()}' did not select an input (tried: '{selector.input_to_select}')"
        )
    use_basename = (
        tool_in.localise_file
        or isinstance(selector, InputSelector)
        and selector.use_basename
    )
    expression = base_expression if not use_basename else f"basename({base_expression})"

    outputs = [wdl.Output(out.output_type.wdl(), out.id(), expression)]
    for s in value_or_default(out.output_type.secondary_files(), []):
        sec_expression = None
        if "^" not in s:
            # do stuff here
            sec_expression = f'{expression} + "{s.replace("^", "")}"'

        elif isinstance(tool_in.input_type, Filename) and tool_in.input_type.extension:
            # use the wdl function: sub
            sec_expression = 'sub({inp}, "\\\\{old_ext}$", "{new_ext}")'.format(
                inp=expression,
                old_ext=tool_in.input_type.extension,
                new_ext=s.replace("^", ""),
            )

        elif (
            File().can_receive_from(tool_in.input_type)
            and isinstance(tool_in.input_type, File)
            and tool_in.input_type.extension
        ):
            # use basename
            sec_expression = f'basename({expression}, "{tool_in.input_type.extension}") + "{s.replace("^", "")}"'

        outputs.append(
            wdl.Output(
                out.output_type.wdl(),
                get_secondary_tag_from_original_tag(out.id(), s),
                sec_expression,
            )
        )
    return outputs


def translate_string_formatter_for_output(
    out, selector: StringFormatter, inp_map: Dict[str, ToolInput], **debugkwargs
) -> List[wdl.Output]:
    """
    The output glob was a string formatter, so we'll need to build the correct glob
    by resolving the string formatter. Some minor complications involve how an output
    with a secondary file must resolve input selectors.

    For example, if you are generating an output with the Filename class, and your output
    type has secondary files, you will need to translate the generated filename with
    respect to the secondary file extension. Or the File class also has a recommended
    "extension" property now that this should consider.

    :param inp_map:
    :param out:
    :param selector:
    :return:
    """
    inputs_to_retranslate = {
        k: v
        for k, v in selector.kwargs.items()
        if not any(isinstance(v, t) for t in StringFormatter.resolved_types)
    }

    has_secondary_files = bool(out.output_type.secondary_files())

    if not has_secondary_files:
        resolved_kwargs = {
            **selector.kwargs,
            **{
                k: get_input_value_from_potential_selector_or_generator(
                    v, inputsdict=inp_map, string_environment=True, **debugkwargs
                )
                for k, v in inputs_to_retranslate.items()
            },
        }

        resolved_exp = selector.resolve_with_resolved_values(**resolved_kwargs)
        return [
            wdl.Output(
                data_type=out.output_type.wdl(),
                name=out.id(),
                expression=f'"{resolved_exp}"',
            )
        ]

    translated_inputs = {}
    input_selectors_with_secondaries = {}

    for k, v in inputs_to_retranslate.items():

        if has_secondary_files and isinstance(v, InputSelector):
            # Handle input selector separately
            input_selectors_with_secondaries[k] = v

        translated_inputs[k] = get_input_value_from_potential_selector_or_generator(
            v, inputsdict=inp_map, string_environment=True, **debugkwargs
        )

    if len(input_selectors_with_secondaries) > 1:
        invalid_keys = ", ".join(input_selectors_with_secondaries.keys())
        raise Exception(
            f"There might be an error when building the string formatter for output '{out.id()}', the "
            f"values to replace for the keys ({invalid_keys}) each which "
            f"had secondary files. This behaviour is currently undefined."
        )
    else:

        inputsel_key = next(iter(input_selectors_with_secondaries.keys()))
        inputsel = input_selectors_with_secondaries[inputsel_key]

        tool_in = inp_map.get(inputsel.input_to_select)

    expression = selector.resolve_with_resolved_values(
        **{**selector.kwargs, **translated_inputs}
    )

    outputs = [wdl.Output(out.output_type.wdl(), out.id(), f'"{expression}"')]
    for s in value_or_default(out.output_type.secondary_files(), []):
        sec_expression = None
        if "^" not in s:
            # do stuff here
            sec_expression = expression + s

        elif tool_in:
            if isinstance(tool_in.input_type, Filename):
                if not tool_in.input_type.extension:
                    raise Exception(
                        f"Unsure how to handle secondary file '{s}' as it uses the escape characater '^' but"
                        f"Janis can't determine the extension of the input '{tool_in.id()} to replace in the "
                        f"WDL translation. You will to annotate the input with an extension Filename(extension=)"
                    )

                # use the wdl function: sub
                sec_expression = '${sub({inp}, "\\\\{old_ext}$", "{new_ext}")}'.format(
                    inp=expression,
                    old_ext=tool_in.input_type.extension,
                    new_ext=s.replace("^", ""),
                )

            elif File().can_receive_from(tool_in.input_type):
                if not tool_in.input_type.extension:
                    raise Exception(
                        f"Unsure how to handle secondary file '{s}' as it uses the escape characater '^' but"
                        f"Janis can't determine the extension of the input '{tool_in.id()} to replace in the "
                        f"WDL translation. When creating an instance of the {tool_in.input_type.__name__} class, "
                        f"it's possible to annotate an extension. For example, the Zip filetype is going to "
                        f'have the .zip extension, so could be initialised with Zip(extension=".zip"'
                    )
                # use basename
                replaced_s = s.replace("^", "")
                sec_expression = f'basename({tool_in.id()}, "{tool_in.input_type.extension}") + "{replaced_s}"'
            else:
                raise Exception(
                    f"Unsure how to handle secondary file '{s}' as it uses the escape characater '^' but"
                    f"Janis can't determine the extension of the input '{tool_in.id()} to replace in the "
                    f"WDL translation. Extra handling in 'wdl.translate_string_formatter_for_output' may be"
                    f"required for the '{tool_in.input_type.__name__}' type might be required."
                )
        else:
            sec_expression = apply_secondary_file_format_to_filename(expression, s)

        outputs.append(
            wdl.Output(
                out.output_type.wdl(),
                get_secondary_tag_from_original_tag(out.id(), s),
                f'"{sec_expression}"',
            )
        )

    return outputs


def translate_step_node(
    node2,
    step_identifier: str,
    resource_overrides: Dict[str, str],
    invalid_identifiers: Set[str],
) -> wdl.WorkflowCallBase:
    """
    Convert a step into a wdl's workflow: call { **input_map }, this handles creating the input map and will
    be able to handle multiple scatters on this step node. If there are multiple scatters, the scatters will be ordered
    in to out by alphabetical order.
    
    This method isn't perfect, when there are multiple sources it's not correctly resolving defaults,
    and tbh it's pretty confusing.
    
    :param node:
    :param step_identifier:
    :param step_alias:
    :param resource_overrides:
    :return:
    """
    from janis_core.workflow.workflow import StepNode, InputNode

    node: StepNode = node2
    step_alias: str = node.id()

    ins = node.inputs()

    # Sanity check our step node connections:

    # 1. make sure our inputs are all present:
    missing_keys = [
        k
        for k in ins.keys()
        if k not in node.sources and not (ins[k].intype.optional or ins[k].default)
    ]
    if missing_keys:
        raise Exception(
            f"Error when building connections for step '{node.id()}', "
            f"missing the required connection(s): '{', '.join(missing_keys)}'"
        )

    # 2. gather the scatters, and make sure none of them are derived from multiple sources, otherwise
    #       we're like double zipping things, it's complicated and it's even MORE complicated in WDL.
    scatterable: List[StepTagInput] = []

    if node.scatter:
        unbound_scatter_keys = [k for k in node.scatter.fields if k not in node.sources]
        if len(unbound_scatter_keys):
            raise Exception(
                f"Attempted to scatter {node.id()} on field(s) [{', '.join(unbound_scatter_keys)}] however "
                "these inputs were not mapped on step construction. Make sure that those unbound keys exist"
                f"in your step definition (eg: "
                f"{node.tool.__class__.__name__}({', '.join(k + '=inp' for k in unbound_scatter_keys)})"
            )
        scatterable = [node.sources[k] for k in node.scatter.fields]

        invalid_sources = [
            si
            for si in scatterable
            if si.multiple_inputs or isinstance(si.source(), list)
        ]
        if len(invalid_sources) > 0:
            invalid_sources_str = ", ".join(
                si.dotted_source() for si in invalid_sources
            )
            raise NotImplementedError(
                f"The edge(s) '{invalid_sources_str}' on node '{node.id()}' scatters"
                f"on multiple inputs, this behaviour has not been implemented"
            )

    # 1. Generate replacement of the scatterable key(s) with some random variable, eg: for i in iterable:
    scattered_old_to_new_identifier = generate_scatterable_details(
        scatterable, forbiddenidentifiers=invalid_identifiers
    )

    # Let's map the inputs, to the source. We're using a dictionary for the map atm, but WDL requires the _format:
    #       fieldName: sourceCall.Output

    inputs_map = {}
    for k in ins:
        if k not in node.sources:
            continue

        array_input_from_single_source = False

        edge: StepTagInput = node.sources[k]
        source: Edge = edge.source()  # potentially single item or array

        # We have multiple sources going to the same entry
        if isinstance(source, list):
            if len(source) == 1:
                source = source[0]
            elif len(source) > 1:

                # There are multiple sources, this is sometimes a little tricky
                input_name_maps = ", ".join(edge.dotted_source())

                multiple_sources_failure_reasons = []

                unique_types = set()
                for x in edge.source():
                    t: TOutput = (
                        first_value(x.start.outputs())
                        if not x.stag
                        else x.start.outputs()[x.stag]
                    )

                    unique_types.update(t.outtype.secondary_files() or [""])
                if len(unique_types) > 1:
                    multiple_sources_failure_reasons.append(
                        f"has {len(unique_types)} different DataTypes with varying secondaries"
                    )
                if node.scatter:
                    multiple_sources_failure_reasons.append(f"is scattered")

                if len(multiple_sources_failure_reasons) > 0:
                    reasons = " and ".join(multiple_sources_failure_reasons)
                    Logger.critical(
                        f"Conversion to WDL for field '{node.id()}.{k}' does not fully support multiple sources."
                        f" This will only work if all of the inputs ({input_name_maps}) have the same secondaries "
                        f"AND this field ('{k}') is not scattered. However this connection {reasons}"
                    )

                ds = edge.dotted_source()
                if isinstance(ds, list):
                    ds = "[" + ", ".join(ds) + "]"
                inputs_map[k] = ds
                f = edge.finish.inputs()[edge.ftag]
                secs = (
                    f.intype.subtype().secondary_files()
                    if isinstance(f.intype, Array)
                    else f.intype.secondary_files()
                )
                if secs:
                    for sec in secs:
                        inputs_map[get_secondary_tag_from_original_tag(k, sec)] = (
                            "["
                            + ", ".join(
                                get_secondary_tag_from_original_tag(kk, sec)
                                for kk in edge.dotted_source()
                            )
                            + "]"
                        )
                continue

        elif source:
            it = source.finish.inputs()[source.ftag].intype
            if source.stag:
                ot = source.start.outputs()[source.stag].outtype
            else:
                ot = first_value(source.start.outputs()).outtype
            if (
                isinstance(it, Array)
                and not isinstance(ot, Array)
                and not source.scatter
            ):
                array_input_from_single_source = True
        secondary = None
        # We're connecting to another step
        if source and isinstance(source.finish, StepNode) and source.ftag:

            it = source.finish.inputs()[source.ftag].intype

            secondary = (
                it.subtype().secondary_files()
                if isinstance(it, Array)
                else it.secondary_files()
            )
            if secondary and isinstance(source.start, StepNode) and source.stag:

                ot = source.start.outputs()[source.stag].outtype

                sec_out = set(
                    value_or_default(
                        ot.subtype().secondary_files()
                        if isinstance(ot, Array)
                        else ot.secondary_files(),
                        default=[],
                    )
                )
                sec_in = set(secondary)
                if not sec_out.issubset(sec_in):
                    raise Exception(
                        f"An error occurred when connecting '{source.source_dotted()}' to "
                        f"'{source.finish.id()}.{source.ftag}', there were secondary files in the final node "
                        f"that weren't present in the source: {', '.join(sec_out.difference(sec_in))}"
                    )

        # Edge defaults
        if not source:
            # edge but no source, probably a default
            if not edge.default:
                Logger.critical(
                    f"Skipping connection to '{edge.finish}.{edge.ftag}' had no source or default, "
                    f"please raise an issue as investigation may be required"
                )
                continue

            if isinstance(edge.default, bool):
                inputs_map[k] = "true" if edge.default else "false"
            elif isinstance(edge.default, str):
                inputs_map[k] = f'"{edge.default}"'
            else:
                inputs_map[k] = edge.default

        # Scattering on multiple secondary files
        elif edge in scatterable and secondary:
            # We're ensured through inheritance and .receiveBy that secondary files will match.
            ds = source.source_dotted()
            Logger.log(
                f"Oh boii, we're gonna have some complicated scattering here with {len(secondary)} secondary file(s)"
            )

            identifier = scattered_old_to_new_identifier[ds]
            inputs_map[k] = identifier[0] + "[0]"
            for idx in range(len(secondary)):
                sec = secondary[idx]
                inputs_map[
                    get_secondary_tag_from_original_tag(k, sec)
                ] = f"{identifier[0]}[{idx + 1}]"

        else:
            ds = source.source_dotted()
            default = None
            if source.start and isinstance(source.start, InputNode):
                default = source.start.default

            inpsourcevalue = None
            if (
                ds in scattered_old_to_new_identifier
                and scattered_old_to_new_identifier[ds]
            ):
                # can't get here with secondary

                s = scattered_old_to_new_identifier[ds]
                inpsourcevalue = s[0]
                default = None

            else:
                inpsourcevalue = ds
                if secondary:
                    if default:
                        Logger.critical(
                            f"The default '{default}' will not be applied to the map '{ds}' as there are secondary files"
                        )
                        default = None
                    for idx in range(len(secondary)):
                        sec = secondary[idx]
                        inputs_map[
                            get_secondary_tag_from_original_tag(k, sec)
                        ] = get_secondary_tag_from_original_tag(ds, sec)

            if default:
                defval = get_input_value_from_potential_selector_or_generator(
                    default, inputsdict=None, string_environment=False, dottedsource=ds
                )

                if isinstance(defval, bool):
                    defval = "true" if defval else "false"

                inpsourcevalue = f"select_first([{inpsourcevalue}, {defval}])"

            if array_input_from_single_source and not (
                isinstance(source.start, StepNode) and source.start.scatter
            ):
                inpsourcevalue = f"[{inpsourcevalue}]"
                if secondary:
                    for sec in secondary:
                        tag = get_secondary_tag_from_original_tag(k, sec)
                        inputs_map[tag] = f"[{inputs_map[tag]}]"

            inputs_map[k] = inpsourcevalue

    inputs_map.update(resource_overrides)

    call = wdl.WorkflowCall(step_identifier, step_alias, inputs_map)

    if len(scatterable) > 0:
        call = wrap_scatter_call(
            call, node.scatter, scatterable, scattered_old_to_new_identifier
        )

    return call


def generate_scatterable_details(
    scatterable: List[StepTagInput], forbiddenidentifiers: Set[str]
):

    # this dictionary is what we're going to use to map our current
    # identifier to the scattered identifier. This step is just the
    # setup, and in the next for loop, we'll
    scattered_old_to_new_identifier = {
        k.dotted_source(): (k.dotted_source(), k.source())
        for k in scatterable
        if not isinstance(k.dotted_source(), list)
    }

    # Make a copy of the forbiddenIds and add the identifiers of the source
    forbiddenidentifierscopy = set(forbiddenidentifiers).union(
        set(v[0] for v in scattered_old_to_new_identifier.values())
    )

    # We'll wrap everything in the scatter block later, but let's replace the fields we need to scatter
    # with the new scatter variable (we'll try to guess one based on the fieldname). We might need to eventually
    # pass the workflow inputs to make sure now conflict will arise.

    if len(scatterable) > 1:
        # We'll generate one variable in place
        standin = generate_new_id_from("Q", forbiddenidentifierscopy)

        # Store the the standin variable for us to use later (as an illegal identifier character)
        scattered_old_to_new_identifier["-"] = standin
        forbiddenidentifierscopy.add(standin)

        # Then we'll need to generate something like:
        #       A -> Q[0], B -> Q[0][0] -> ..., N -> Q([0] * (n-1))[1]
        n = len(scatterable)
        for i in range(len(scatterable)):
            s = scatterable[i]
            newid = standin + (i) * ".right" + ".left"
            if i == n - 1:
                newid = standin + (n - 1) * ".right"
            scattered_old_to_new_identifier[s.dotted_source()] = (
                newid,
                first_value(s.source_map).start,
            )
            forbiddenidentifierscopy.add(newid)
    else:

        for s in scatterable:

            # We asserted earlier that the source_map only has one value (through multipleInputs)
            e: Edge = first_value(s.source_map)
            newid = generate_new_id_from(
                e.stag or e.source_dotted().replace(".", ""), forbiddenidentifierscopy
            )
            scattered_old_to_new_identifier[s.dotted_source()] = (newid, e.start)
            forbiddenidentifierscopy.add(newid)

    return scattered_old_to_new_identifier


def wrap_scatter_call(
    call, scatter: ScatterDescription, scatterable, scattered_old_to_new_identifier
):
    from janis_core.workflow.workflow import StepNode, InputNode

    # Let's start the difficult process of scattering, in WDL we'll:
    #
    #       1. We already generated the new "scatter" variable that will
    #            be used in place of the original .dotted_source()
    #
    #       2. Generate annotations for accessory / secondary files
    #
    #       3. Wrap everything in the appropriate scatter block,
    #           especially considering scattering by multiple variables
    #

    # 2. Explanation:
    #     So, the current way of mixing accessory files is not really                        _
    #     supported, but a little complicated basically, if our scatterable edge           _| |
    #     contains secondary files, they'll all be arrays of separate files, eg:         _| | |
    #                                                                                   | | | | __
    #     File[] bams =     [...]                                                       | | | |/  \
    #     File[] bams_bai = [...]                                                       |       /\ \
    #                                                                                   |       \/ /
    #     We can handle this by transposing the array of both items, eg:                 \        /
    #                                                                                     |      /
    #         transpose([bam1, bam2, ..., bamn], [bai1, bai2, ..., bai3])                 |     |
    #               => [[bam1, bai1], [bam2, bai2], ..., [bamn, bain]]
    #
    #     and then unwrap them using their indices and hopefully have everything line up:
    #
    #     Source: https://software.broadinstitute.org/wdl/documentation/spec#arrayarrayx-transposearrayarrayx

    # We need to generate the source statement, this is easy if we're scattering by one variable

    # sanity check
    if len(scatterable) == 0:
        return call

    # generate the new source map

    insource_ar = []
    for s in scatterable:
        secondary = s.finish.tool.inputs_map()[s.ftag].intype.secondary_files()
        if secondary:
            ds = s.dotted_source()
            joined_tags = ", ".join(
                get_secondary_tag_from_original_tag(ds, sec) for sec in secondary
            )
            transformed = f"transpose([{ds}, {joined_tags}])"
            insource_ar.append(transformed)

        else:
            (newid, startnode) = scattered_old_to_new_identifier[s.dotted_source()]
            insource = s.dotted_source()
            if isinstance(startnode, InputNode) and startnode.default is not None:
                resolved = get_input_value_from_potential_selector_or_generator(
                    startnode.default,
                    None,
                    string_environment=False,
                    scatterstep=insource,
                )
                if isinstance(resolved, bool):
                    resolved = "true" if resolved else "false"

                insource_ar.append(f"select_first([{insource}, {resolved}])")
            else:
                insource_ar.append(insource)

    insource = None
    alias = None
    if len(insource_ar) == 1:
        insource = insource_ar[0]
        alias = first_value(scattered_old_to_new_identifier)[0]
    else:
        method = "zip" if scatter.method == ScatterMethods.dot else "cross"
        insource = recursive_2param_wrap(method, insource_ar)
        alias = scattered_old_to_new_identifier["-"]

    return wdl.WorkflowScatter(alias, insource, [call])


## SELECTOR HELPERS


def get_input_value_from_potential_selector_or_generator(
    value, inputsdict, string_environment=True, **debugkwargs
):
    """
    We have a value which could be anything, and we want to convert it to a expressionable value.
    If we have a string, it should be in quotes, etc. It should be "Type paramname = <expressionable>"
    :param value:
    :param tool_id:
    :param string_environment:  Do we need to wrap string literals in quotes,
                                or do we need to wrap variables in expr block.
    :return:
    """
    if value is None:
        return None

    if isinstance(value, list):
        toolid = value_or_default(debugkwargs.get("tool_id"), "get-value-list")
        joined_values = ", ".join(
            str(
                get_input_value_from_potential_selector_or_generator(
                    value[i],
                    inputsdict,
                    string_environment=False,
                    tool_id=toolid + "." + str(i),
                )
            )
            for i in range(len(value))
        )
        return f"[{joined_values}]"
    elif isinstance(value, str):
        return value if string_environment else f'"{value}"'
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, int) or isinstance(value, float):
        return value
    elif isinstance(value, Filename):
        return (
            value.generated_filename()
            if string_environment
            else f'"{value.generated_filename()}"'
        )
    elif isinstance(value, StringFormatter):
        return translate_string_formatter(
            selector=value,
            inputsdict=inputsdict,
            string_environment=string_environment,
            **debugkwargs,
        )
    elif isinstance(value, WildcardSelector):
        raise Exception(
            f"A wildcard selector cannot be used as an argument value for '{debugkwargs}'"
        )
    # elif isinstance(value, CpuSelector):
    #     return translate_cpu_selector(
    #         value, inputsdict, string_environment=string_environment
    #     )
    # elif isinstance(value, MemorySelector):
    #     return translate_mem_selector(
    #         value, inputsdict, string_environment=string_environment
    #     )
    elif isinstance(value, InputSelector):
        return translate_input_selector(
            selector=value,
            inputsdict=inputsdict,
            string_environment=string_environment,
            **debugkwargs,
        )
    elif callable(getattr(value, "wdl", None)):
        return value.wdl()

    warning = ""
    if isclass(value):
        stype = value.__name__
        warning = f", this is likely due to the '{stype}' not being initialised"
    else:
        stype = value.__class__.__name__
    raise Exception(
        f"Could not detect type '{stype}' to convert to input value{warning}"
    )


def translate_string_formatter(
    selector: StringFormatter, inputsdict, string_environment=False, **debugkwargs
):
    # we should raise an Exception if any of our inputs are optional without a default

    invalid_select_inputs = [
        (k, selector.kwargs[k].input_to_select)
        for k in selector.kwargs
        # Our selector is getting an input
        if isinstance(selector.kwargs[k], InputSelector)
        and selector.kwargs[k].input_to_select in inputsdict
        and not isinstance(
            inputsdict[selector.kwargs[k].input_to_select].input_type, Filename
        )
        # our selected input is optional
        and inputsdict[selector.kwargs[k].input_to_select].input_type.optional
        # our selected input does NOT have a default
        # tbh, this ToolInput might have a default that selects a different input that is null,
        # but I'm not going down this rabbit hole
        and inputsdict[selector.kwargs[k].input_to_select].default is None
    ]

    if len(invalid_select_inputs) > 0:
        tags = ", ".join(f"'{k[0]}'" for k in invalid_select_inputs)
        inps = ", ".join(f"'{k[1]}'" for k in invalid_select_inputs)
        raise Exception(
            f'There was an error when resolving the format "{selector._format}", the tag(s) {tags} respectively '
            f"selected input(s) {inps} that were optional and did NOT have a default value."
        )

    value = selector.resolve_with_resolved_values(
        **{
            k: get_input_value_from_potential_selector_or_generator(
                selector.kwargs[k],
                inputsdict=inputsdict,
                string_environment=True,
                **debugkwargs,
            )
            for k in selector.kwargs
        }
    )

    return value if string_environment else f'"{value}"'


def translate_input_selector(
    selector: InputSelector, inputsdict, string_environment=True, **debugkwargs
):
    if not selector.input_to_select:
        raise Exception("No input was selected for input selector: " + str(selector))

    if not inputsdict:
        raise Exception(
            f"An internal error has occurred when selecting the input '{selector.input_to_select}'"
        )

    if selector.input_to_select not in inputsdict:
        raise Exception(
            f"Couldn't find input '{selector.input_to_select}' in tool '{debugkwargs}'"
        )

    inp = inputsdict[selector.input_to_select]
    name = resolve_tool_input_value(inp, **debugkwargs)
    if string_environment:
        return f"${{{name}}}"
    else:
        return name


# These are now handled entirely by the translate_input_selector_class

# def translate_cpu_selector(selector: CpuSelector, inputsdict, string_environment=True):
#     return translate_input_selector(selector, inputsdict, string_environment=string_environment)
# def translate_mem_selector(selector: MemorySelector, inputsdict, string_environment=True):
#     return translate_input_selector(selector, inputsdict, string_environment=string_environment)


def translate_wildcard_selector(selector: WildcardSelector):
    if not selector.wildcard:
        raise Exception(
            "No wildcard was selected for wildcard selector: " + str(selector)
        )
    return f'glob("{selector.wildcard}")'


## HELPER METHODS


def value_or_default(ar, default):
    """
    default is ar is None, else return the value of ar, returns ar even if ar is FALSEY
    :param ar:
    :param default:
    :return:
    """
    return default if ar is None else ar


def build_aliases(steps2):
    """
    From a list of stepNodes, generate the toolname alias (tool name to unique import alias)
    and the step alias, which
    :param steps: list of step nodes
    :return:
    """
    from janis_core.workflow.workflow import StepNode

    steps: List[StepNode] = steps2

    get_alias = lambda t: t[0] + "".join([c for c in t[1:] if c.isupper()])
    aliases: Set[str] = set()

    tools: List[Tool] = [s.tool for s in steps]
    tool_name_to_tool: Dict[str, Tool] = {t.id().lower(): t for t in tools}
    tool_name_to_alias = {}
    steps_to_alias: Dict[str, str] = {
        s.id().lower(): get_alias(s.id()).lower() for s in steps
    }

    for tool in tool_name_to_tool:
        a = get_alias(tool).upper()
        s = a
        idx = 2
        while s in aliases:
            s = a + str(idx)
            idx += 1
        aliases.add(s)
        tool_name_to_alias[tool] = s

    return tool_name_to_alias, steps_to_alias


def get_secondary_tag_from_original_tag(original, secondary):
    secondary_without_punctuation = secondary.replace(".", "").replace("^", "")
    return original + "_" + secondary_without_punctuation


def apply_secondary_file_format_to_filename(
    filepath: Optional[str], secondary_file: str
):
    """
    This is actually clever, you can probably trust this to do what you want.
    :param filepath: Filename to base
    :param secondary_file: CWL secondary format (Remove 1 extension for each leading ^.
    """
    if not filepath:
        return None

    fixed_sec = secondary_file.lstrip("^")
    leading = len(secondary_file) - len(fixed_sec)
    if leading <= 0:
        return filepath + fixed_sec

    basepath = ""
    filename = filepath
    if "/" in filename:
        idx = len(filepath) - filepath[::-1].index("/")
        basepath = filepath[:idx]
        filename = filepath[idx:]

    split = filename.split(".")

    newfname = filename + fixed_sec
    if len(split) > 1:
        newfname = ".".join(split[: -min(leading, len(split) - 1)]) + fixed_sec
    return basepath + newfname


def prepare_move_statement_for_potential_secondaries(ti: ToolInput):
    it = ti.input_type
    if not issubclass(type(it), File) or not it.secondary_files():
        return []

    commands = []
    for s in it.secondary_files():
        sectag = get_secondary_tag_from_original_tag(ti.id(), s)
        commands.append(
            wdl.Task.Command(
                f'if [ $(dirname "${{{sectag}}}") != $(dirname "{ti.id()}") ]; then mv ${{{sectag}}} $(dirname ${{{ti.id()}}}); fi'
            )
        )
    return commands


def prepare_env_var_setters(
    reqs: Dict[str, Any], inputsdict, **debugkwargs
) -> List[wdl.Task.Command]:
    if not reqs:
        return []

    statements = []
    for k, v in reqs.items():
        val = get_input_value_from_potential_selector_or_generator(
            v, inputsdict=inputsdict, string_environment=True, **debugkwargs
        )
        statements.append(wdl.Task.Command(f"export {k}='{val}'"))

    return statements


def prepare_move_statement_for_input_to_localise(ti: ToolInput):
    it = ti.input_type

    if issubclass(type(it), File):
        commands = [wdl.Task.Command(f"mv ${{{ti.id()}}} -t .")]
        if it.secondary_files():
            for s in it.secondary_files():
                commands.append(
                    wdl.Task.Command(
                        f"mv ${{{get_secondary_tag_from_original_tag(ti.id(), s)}}} -t ."
                    )
                )
        return commands
    if isinstance(it, Array) and issubclass(type(it.subtype()), File):
        subtype = it.subtype()
        commands = [wdl.Task.Command("mv ${{sep=' ' {s}}} -t .".format(s=ti.id()))]
        if subtype.secondary_files():
            for s in it.secondary_files():
                commands.append(
                    wdl.Task.Command(
                        "mv ${{sep=' ' {s}}} -t .".format(
                            s=get_secondary_tag_from_original_tag(ti.id(), s)
                        )
                    )
                )

        return commands

    raise Exception(f"WDL is unable to localise type '{type(it)}'")


def build_resource_override_maps_for_tool(tool, prefix=None) -> List[wdl.Input]:
    inputs = []

    if not prefix:
        prefix = ""  # wf.id() + "."
    else:
        prefix += "_"

    if isinstance(tool, CommandTool):
        inputs.extend(
            [
                wdl.Input(wdl.WdlType.parse_type("Int?"), prefix + "runtime_memory"),
                wdl.Input(wdl.WdlType.parse_type("Int?"), prefix + "runtime_cpu"),
                wdl.Input(wdl.WdlType.parse_type("String"), prefix + "runtime_disks"),
            ]
        )
    else:
        inputs.extend(build_resource_override_maps_for_workflow(tool, prefix=prefix))

    return inputs


def build_resource_override_maps_for_workflow(wf, prefix=None) -> List[wdl.Input]:

    # returns a list of key, value pairs
    inputs = []

    for s in wf.step_nodes.values():
        tool: Tool = s.tool

        tool_pre = (prefix or "") + s.id()
        inputs.extend(build_resource_override_maps_for_tool(tool, prefix=tool_pre))

    return inputs



"""
    node.py

    Provides base class that different nodes must override, this translates closest to a Step
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Any

from janis_core.tool.tool import TInput, TOutput

NodeLabel = str
NodeType = int


class NodeTypes:
    INPUT: NodeType = 1
    OUTPUT: NodeType = 2
    STEP: NodeType = 3

    @staticmethod
    def to_str(node_type: NodeType) -> str:
        if node_type == NodeTypes.INPUT:
            return "Input"
        if node_type == NodeTypes.OUTPUT:
            return "Output"
        if node_type == NodeTypes.STEP:
            return "Task"
        raise Exception(f"Unhandled task type: '{node_type}'")

    @staticmethod
    def to_col(node_type: NodeType) -> str:
        if node_type == NodeTypes.INPUT:
            return "red"
        if node_type == NodeTypes.OUTPUT:
            return "lightblue"
        if node_type == NodeTypes.STEP:
            return "blue"
        raise Exception(f"Unhandled task type: '{node_type}'")


class Node(ABC):

    _N_counter: int = 1
    _N_nodeId_map: Dict[int, Any] = {}

    def __init__(self, wf, node_type: NodeType, identifier: NodeLabel, depth=0):

        self.wf = wf
        self.node_type: NodeType = node_type
        self.identifier: NodeLabel = identifier
        self.depth = depth

        # actually a StepInput
        self.sources: Dict[str, any] = {}

        # Update unique counter for hash
        self._nodeId = Node._N_counter
        Node._N_counter += 1

        # Map the node, so we can look it up later
        self._N_nodeId_map[self._nodeId] = self

    def id(self) -> str:
        return self.identifier

    def __hash__(self):
        return self._nodeId

    def __repr__(self):
        return f"{self.node_type}: {self.id()}"

    def __str__(self):
        return f"{NodeTypes.to_str(self.node_type)}: {self.identifier}"

    def set_depth(self, depth: int):
        self.depth = max(self.depth, depth)

    def __eq__(self, other):
        if isinstance(other, Node):
            return self._nodeId == other._nodeId
        return False

    @abstractmethod
    def inputs(self) -> Dict[str, TInput]:
        raise Exception(
            f"Subclass {type(self)} must implement inputs, return dict: key: ToolInput"
        )

    @abstractmethod
    def outputs(self) -> Dict[str, TOutput]:
        raise Exception(
            f"Subclass {type(self)} must implement outputs, return dict: key: ToolOutput"
        )


NodeAndTag = Tuple[str, Node]


def layout_nodes(nodes: List[Node], n_inputs: int = 0) -> Dict[Node, Tuple[int, int]]:
    """
    Stack on depth away from root, and scale smaller columns
    :param n_inputs:
    :param nodes: list of nodes in Graph to position
    :return: Dict[Node, (x,y)]
    """
    pos = {}
    # int is the depth at the specified index
    cur_depth: List[int] = []
    depth_node: Dict[int, List[Node]] = {}

    def get_idx_for_depth(d: int):
        d_idx = d
        m = len(cur_depth)
        if d_idx >= m:
            cur_depth.extend([0] * (d - m + 1))
        return cur_depth[d]

    def push_to_dict(key, val):
        if key in depth_node:
            depth_node[key].append(val)
        else:
            depth_node[key] = [val]

    for node in nodes:
        depth = node.depth
        push_to_dict(depth, node)
        d = get_idx_for_depth(depth)
        cur_depth[depth] += 1
        pos[node] = (depth, d)

    # Now normalise each depth
    max_in_col = float(max(cur_depth + [n_inputs]))

    for (idx, d) in enumerate(cur_depth):
        if idx not in depth_node:
            continue
        nodes = depth_node[idx]
        scale = max_in_col / len(nodes)

        # If max and this col differ in even / odd, add half a unit
        half_bias = 0 if (len(nodes) % 2) == (max_in_col % 2) else 0.5

        for node in nodes:
            (x, y) = pos[node]
            pos[node] = (x, (y + half_bias) * scale)

    return pos


def layout_nodes2(nodes: List[Node]) -> Dict[Node, Tuple[int, int]]:
    # Aim for something like Rabix
    inputs = [n for n in nodes if n.node_type == NodeTypes.INPUT]
    others = [n for n in nodes if n.node_type != NodeTypes.INPUT]

    pos = layout_nodes(others, len(inputs))
    s = 0
    for n in inputs:
        pos[n] = (0, s)
        s += 1
    return pos


def layout_nodes3(nodes: List[Node]) -> Dict[Node, Tuple[int, int]]:
    # Other ideas to scale might be, start with steps and
    # stack in a similar way to depth around center, then do the same
    return {}



from uuid import uuid4

scattered_ordered_variable_identifiers = [
    "i",
    "j",
    "k",
    "x",
    "y",
    "z",
    "a",
    "b",
    "c",
    "ii",
    "jj",
    "kk",
    "xx",
    "yy",
    "zz",
]


def generate_new_id_from(source, forbidden_ids):
    new_var = source[0]
    max_size_of_subid = None
    while new_var in forbidden_ids:
        if len(scattered_ordered_variable_identifiers) > 0:
            new_var = scattered_ordered_variable_identifiers.pop(0)
        elif max_size_of_subid == len(source):
            # generate guid
            new_var = str(uuid4()).replace("-", "_")
        else:
            max_size_of_subid = (max_size_of_subid or 0) + 1
            new_var = source[-max_size_of_subid:]
    return new_var



#
# Utils to help out yaml dump
#


# Provide better formatting of multiline literals
# from: https://github.com/cloudlinux/kuberdock-platform/blob/master/kubedock/kapi/apps.py
def str_presenter(dumper, data):
    # check for multiline strings
    if len(data.splitlines()) == 1 and data[-1] == "\n":
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=">")
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data.strip())



from enum import Enum

SupportedTranslation = str


class SupportedTranslations(Enum):
    CWL = "cwl"
    WDL = "wdl"

    def __str__(self):
        return self.value

    def __eq__(self, other):
        return str(self) == str(other)

    @staticmethod
    def all():
        return [SupportedTranslations.CWL, SupportedTranslations.WDL]
